"""
Created on 24.05.2016

@author: Sebastian Illing
"""
import os


def test_solr_search(dummy_solr):
    # search some files
    from evaluation_system.model.solr import SolrFindFiles

    solr_search = SolrFindFiles()
    all_files = solr_search.search()
    assert len(list(all_files)) == 3
    hist = solr_search.search(experiment="historical")
    assert list(hist) == [os.path.join(dummy_solr.tmpdir, dummy_solr.files[0])]
    all_files = solr_search.search(latest_version=False)
    assert len(list(all_files)) == 5
    # test OR query
    or_result = solr_search.search(variable=["tauu", "wetso2"])
    assert set(
        [os.path.join(dummy_solr.tmpdir, e) for e in dummy_solr.files[:2]]
    ) == set(or_result)


def test_facet_search(dummy_solr):

    from evaluation_system.model.solr import SolrFindFiles

    factes_to_be = {
        "cmor_table": ["aero", 1, "amon", 2],
        "product": ["output1", 3],
        "realm": ["aerosol", 1, "atmos", 2],
        "dataset": ["cmip5", 3],
        "institute": ["mohc", 3],
        "project": ["cmip5", 3],
        "time_frequency": ["mon", 3],
        "experiment": ["decadal2008", 1, "decadal2009", 1, "historical", 1],
        "variable": ["tauu", 1, "ua", 1, "wetso2", 1],
        "model": ["hadcm3", 3],
        "ensemble": ["r2i1p1", 1, "r7i2p1", 1, "r9i3p1", 1],
    }
    s = SolrFindFiles
    all_factes = s.facets()
    assert len(all_factes) == 11
    assert all_factes == factes_to_be

    var_facets = s.facets(facets=["variable"])
    assert var_facets == dict(variable=factes_to_be["variable"])
    experiment_facets = s.facets(facets="experiment", cmor_table="amon")
    assert experiment_facets == {"experiment": ["decadal2008", 1, "decadal2009", 1]}

    # test files core
    res = s.facets(facets="variable,project", latest_version=False)
    assert list(res.keys()) == ["variable", "project"]
    assert res == {
        "variable": ["tauu", 1, "ua", 3, "wetso2", 1],
        "project": ["cmip5", 5],
    }



from django.conf import settings
from evaluation_system.misc import config

import json, time
import django

SETTINGS = dict(TIME_ZONE="UTC", USE_TZ=False)
config.reloadConfiguration()
# Application definition
SETTINGS["INSTALLED_APPS"] = (
    "django.contrib.auth",  # We need this to access user groups
    "django.contrib.flatpages",
    "django.contrib.contenttypes",
    "django.contrib.sites",
)
SETTINGS["DATABASES"] = {
    "default": {
        "ENGINE": "django.db.backends.mysql",
        "NAME": config.get(config.DB_DB),
        "USER": config.get(config.DB_USER),
        "PASSWORD": config.get(config.DB_PASSWD),
        "HOST": config.get(
            config.DB_HOST
        ),  # Or an IP Address that your DB is hosted on
        "PORT": config.get(str(config.DB_PORT), "3306"),
    }
}
try:
    settings.configure(**SETTINGS)
    django.setup()
except RuntimeError:
    pass



TEST_EVAL = """[evaluation_system]
base_dir=evaluation_system
project_name=freva-ces
base_dir_location=/tmp/evaluation_system_test
scratch_dir=/tmp/scratch/$$USER
directory_structure_type=central
scheduler_input_dir=/tmp/slurm
scheduler_output_dir=/tmp/slurm
scheduler_system=slurm
preview_path=/tmp/preview
project_data=/tmp/crawl_my_data
db.host=127.0.0.1
db.user=freva
db.passwd=T3st
db.db=freva
solr.host=localhost
solr.port=8983
solr.core=files
[scheduler_options]
source=$$EVALUATION_SYSTEM_HOME/activate_sh
queue=compute
project=ch1187
memory=10G
[plugin:DummyPlugin]
plugin_path=$$EVALUATION_SYSTEM_HOME/src/evaluation_system/tests/mocks
python_path=$$EVALUATION_SYSTEM_HOME/src/evaluation_system/tests/mocks
module=dummy"""

TEST_DRS = """
[cmip5]
root_dir = "/mnt/data4freva/model/global"
parts_dir = [ "project", "product", "institute", "model", "experiment", "time_frequency",
    "realm", "cmor_table", "ensemble", "version", "variable",
]
parts_file_name = [ "variable", "cmor_table", "model", "experiment", "ensemble", "time", ]
parts_time = "start_time-end_time"

[cmip5.defaults]
project = "cmip5"

[observations]
root_dir = ".docker/data"
parts_dir = [
    "project", "product", "institute", "model", "experiment", "time_frequency",
    "realm", "cmor_table", "ensemble", "version", "variable",
]
parts_versioned_dataset = [
    "project", "product", "institute", "model", "experiment", "time_frequency",
    "realm", "cmor_table", "ensemble", "version", "variable",
]
parts_file_name = ["variable", "experiment", "level", "version", "time"]
parts_time = "start_time-end_time"

[observations.defaults]
project = "observations"

[reanalysis]
root_dir = "/mnt/data4freva"
parts_dir = [
    "project", "product", "institute", "model", "experiment", "time_frequency",
    "realm", "variable", "ensemble",
]
parts_file_name = [
    "variable", "cmor_table", "project", "experiment", "ensemble", "time",
]
parts_time = "start_time-end_time"

[reanalysis.defaults]
project = "reanalysis"
product = "reanalysis"

[crawl_my_data]
root_dir = "/tmp/crawl_my_data"
parts_dir = [
    "project", "product", "institute", "model", "experiment", "time_frequency",
    "realm", "cmor_table", "ensemble", "variable"
]
parts_file_name = [
    "variable", "cmor_table", "model", "experiment", "ensemble", "time",
]
parts_time = "start_time-end_time"
"""



from __future__ import annotations
from contextlib import contextmanager, suppress
import os
from pathlib import Path
import re
import shlex
import subprocess
import sys
import string
import tempfile
import abc
from typing import Any, Iterator, ClassVar, Optional, Union
from evaluation_system.misc import logger


def parse_bytes(s: Union[str, float, int]) -> int:
    """Parse byte string to numbers

    >>> from dask.utils import parse_bytes
    >>> parse_bytes('100')
    100
    >>> parse_bytes('100 MB')
    100000000
    >>> parse_bytes('100M')
    100000000
    >>> parse_bytes('5kB')
    5000
    >>> parse_bytes('5.4 kB')
    5400
    >>> parse_bytes('1kiB')
    1024
    >>> parse_bytes('1e6')
    1000000
    >>> parse_bytes('1e6 kB')
    1000000000
    >>> parse_bytes('MB')
    1000000
    >>> parse_bytes(123)
    123
    >>> parse_bytes('5 foos')
    Traceback (most recent call last):
        ...
    ValueError: Could not interpret 'foos' as a byte unit
    """
    if isinstance(s, (int, float)):
        return int(s)
    s = s.replace(" ", "")
    if not any(char.isdigit() for char in s):
        s = "1" + s

    for i in range(len(s) - 1, -1, -1):
        if not s[i].isalpha():
            break
    index = i + 1

    prefix = s[:index]
    suffix = s[index:]

    try:
        n = float(prefix)
    except ValueError as e:
        raise ValueError("Could not interpret '%s' as a number" % prefix) from e
    byte_sizes = {
        "kB": 10**3,
        "MB": 10**6,
        "GB": 10**9,
        "TB": 10**12,
        "PB": 10**15,
        "KiB": 2**10,
        "MiB": 2**20,
        "GiB": 2**30,
        "TiB": 2**40,
        "PiB": 2**50,
        "B": 1,
        "": 1,
    }
    byte_sizes = {k.lower(): v for k, v in byte_sizes.items()}
    byte_sizes.update({k[0]: v for k, v in byte_sizes.items() if k and "i" not in k})
    byte_sizes.update({k[:-1]: v for k, v in byte_sizes.items() if k and "i" in k})

    try:
        multiplier = byte_sizes[suffix.lower()]
    except KeyError as e:
        raise ValueError("Could not interpret '%s' as a byte unit" % suffix) from e

    result = n * multiplier
    return int(result)


@contextmanager
def tmpfile(**kwargs) -> Iterator[Path]:
    """
    Function to create and return a unique temporary file with the given
    extension, if provided.

    Parameters
    ----------
    kwargs :
        Keyword arguments to be passed to tempfile.NamedTemporaryFile

    Returns
    -------
    out : Path
        Path to the temporary file

    See Also
    --------
    NamedTemporaryFile : Built-in alternative for creating temporary files
    tmp_path : pytest fixture for creating a temporary directory unique to
               the test invocation
    """
    with tempfile.NamedTemporaryFile(**kwargs) as tf:
        yield Path(tf.name)


def string_to_bytes(size: str) -> float:
    """Convert a size string to int."""

    mul_dec = dict(pb=1000**5, tb=1000**4, gb=1000**3, mb=1000**2, kb=1000, b=1)
    mul_bi = dict(
        pib=1024**5,
        tib=1024**4,
        gib=1024**3,
        mib=1024**2,
        kib=1024,
    )
    num = ""
    byte_form = ""
    for s in str(size).lower().strip():
        if s:
            if s not in string.ascii_lowercase:
                num += s
            else:
                byte_form += s
    if len(byte_form) == 1:
        byte_form += "b"
    try:
        return float(num) * mul_dec[byte_form]
    except KeyError:
        try:
            return float(num) * mul_bi[byte_form]
        except KeyError:
            return float(num)


def format_bytes(n: int) -> str:
    """Format bytes as text

    >>> from dask.utils import format_bytes
    >>> format_bytes(1)
    '1 B'
    >>> format_bytes(1234)
    '1.21 kiB'
    >>> format_bytes(12345678)
    '11.77 MiB'
    >>> format_bytes(1234567890)
    '1.15 GiB'
    >>> format_bytes(1234567890000)
    '1.12 TiB'
    >>> format_bytes(1234567890000000)
    '1.10 PiB'

    For all values < 2**60, the output is always <= 10 characters.
    """
    for prefix, k in (
        ("Pi", 2**50),
        ("Ti", 2**40),
        ("Gi", 2**30),
        ("Mi", 2**20),
        ("ki", 2**10),
    ):
        if n >= k * 0.9:
            return f"{n / k:.2f} {prefix}B"
    return f"{n} B"


class Job(abc.ABC):
    """Base class to launch workers on Job queues

    This class should not be used directly, use a class appropriate for
    your queueing system (e.g. PBScluster or SLURMCluster) instead.

    Parameters
    ----------

    Attributes
    ----------
    submit_command: str
        Abstract attribute for job scheduler submit command,
        should be overridden
    cancel_command: str
        Abstract attribute for job scheduler cancel command,
        should be overridden

    See Also
    --------
    PBSCluster
    SLURMCluster
    SGECluster
    OARCluster
    LSFCluster
    MoabCluster
    """

    _script_template = (
        "%(shebang)s\n\n%(job_header)s\n%(env_header)s\n\n%(worker_command)s"
    )

    # Following class attributes should be overridden by extending classes.
    submit_command: ClassVar[str] = ""
    cancel_command: ClassVar[str] = ""
    config_name: ClassVar[Optional[str]] = None
    job_id_regexp: ClassVar[str] = r"(?P<job_id>\d+)"

    @abc.abstractmethod
    def __init__(
        self,
        name: Optional[str] = None,
        processes: Optional[int] = None,
        local_directory: Optional[Union[str, Path]] = None,
        env_extra: Optional[list[str]] = None,
        header_skip: Optional[list[str]] = None,
        log_directory: Optional[Union[str, Path]] = None,
        memory: Optional[Union[int, str]] = None,
        job_cpu: int = 1,
        cores: int = 2,
        shebang: str = "#!/usr/bin/env bash",
        python: str = sys.executable,
        scheduler: Optional[str] = "freva-plugin",
        freva_args: Optional[list[str]] = None,
        delete_job_script: bool = True,
        **kwargs,
    ):
        self.job_id = ""
        self.memory: int = 0
        env_extra = env_extra or []
        freva_args = freva_args or []
        super().__init__()
        self.job_header = ""
        # Keep information on process, cores, and memory, for use in subclasses
        try:
            self.worker_memory = string_to_bytes(str(memory))
        except ValueError:
            raise ValueError(f"Could not set scheduler memory {str(memory)}")
        self.delete_job_script = delete_job_script
        self.scheduler = scheduler or "freva-plugin"
        self.scheduler += " "
        self.worker_processes = processes
        self.worker_cores = cores or 2
        self.name = self.job_name = name or "worker"
        self.shebang = shebang
        self._env_header = "\n".join(filter(None, env_extra))
        self.header_skip = set(header_skip or [])
        self.log_directory = log_directory
        if self.log_directory is not None:
            if not os.path.exists(self.log_directory):
                os.makedirs(self.log_directory)
        self._command_template = self.scheduler + " ".join(map(str, freva_args))

    def job_script(self) -> str:
        """Construct a job submission script"""
        header = "\n".join(
            [
                line
                for line in self.job_header.split("\n")
                if not any(skip in line for skip in self.header_skip)
            ]
        )
        pieces = {
            "shebang": self.shebang,
            "job_header": header,
            "env_header": self._env_header,
            "worker_command": self._command_template,
        }
        return self._script_template % pieces

    @contextmanager
    def job_file(self) -> Iterator[Path]:
        """Write job submission script to temporary file"""
        with tmpfile(suffix=".sh", delete=self.delete_job_script) as fn:
            with open(fn, "w") as f:
                logger.debug("writing job script: \n%s", self.job_script())
                f.write(self.job_script())
            yield Path(fn)

    def _submit_job(self, script_filename: Union[str, Path]) -> Any:
        # Should we make this async friendly?
        return self._call(shlex.split(self.submit_command) + [str(script_filename)])

    def start(self) -> None:
        """Start workers and point them to our local scheduler"""
        logger.debug("Starting worker: %s", self.name)

        with self.job_file() as fn:
            out = self._submit_job(fn)
            job_id = self._job_id_from_submit_output(out)
            try:
                self.job_id = job_id
            except ValueError:
                self.job_id = "0"
        logger.debug("Starting job: %s", self.job_id)

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def _job_id_from_submit_output(self, out: str) -> str:
        match = re.search(self.job_id_regexp, out)
        if match is None:
            msg = (
                "Could not parse job id from submission command "
                "output.\nJob id regexp is {!r}\nSubmission command "
                "output is:\n{}".format(self.job_id_regexp, out)
            )
            raise ValueError(msg)

        job_id = match.groupdict().get("job_id")
        if job_id is None:
            msg = (
                "You need to use a 'job_id' named group in your regexp, e.g. "
                "r'(?P<job_id>\\d+)'. Your regexp was: "
                "{!r}".format(self.job_id_regexp)
            )
            raise ValueError(msg)

        return job_id

    def close(self) -> None:
        logger.debug("Stopping worker: %s job: %s", self.name, self.job_id)
        self._close_job(self.job_id, self.cancel_command)

    @classmethod
    def _close_job(cls, job_id: str, cancel_command: str) -> None:
        if job_id:
            with suppress(RuntimeError):  # deleting job when job already gone
                cls._call(shlex.split(cancel_command) + [job_id])
            logger.debug("Closed job %s", job_id)

    @staticmethod
    def _call(command: list[str], **kwargs) -> str:
        """Call a command using subprocess.Popen.

        This centralizes calls out to the command line, providing consistent
        outputs, logging, and an opportunity to go asynchronous in the future.

        Parameters
        ----------
        command: List[str]
            A command, each of which is a list of strings to hand to
            subprocess.Popen

        Returns
        -------
        The stdout produced by the command, as string.

        Raises
        ------
        RuntimeError if the command exits with a non-zero exit code
        """
        cmd_str = " ".join(command)
        logger.debug(
            "Executing the following command to command line\n{}".format(cmd_str)
        )
        proc = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs
        )

        out, err = proc.communicate()
        out, err = out.decode(), err.decode()

        if proc.returncode != 0:
            raise RuntimeError(
                "Command exited with non-zero exit code.\n"
                "Exit code: {}\n"
                "Command:\n{}\n"
                "stdout:\n{}\n"
                "stderr:\n{}\n".format(proc.returncode, cmd_str, out, err)
            )
        return out



"""Definitions of parameters types to configure custom ``Freva`` plugins.

Plugin parameters are defined in the plugin wrapper class. Please refer to
:class:`evaluation_system.api.plugin` for more information on how to set up a
plugin wrapper class.

"""
from __future__ import annotations
from collections import defaultdict
import json
import re
import textwrap
from typing import Any, Optional, Union, Type
import warnings

from evaluation_system.misc.utils import find_similar_words, PrintableList, initOrder
from evaluation_system.misc import config
from evaluation_system.misc.exceptions import ValidationError, deprecated_method
from evaluation_system.model.plugins.models import Parameter

ParameterBaseType = Union[str, int, float, bool, PrintableList]
"""Type definitions of all possible parameter types."""


class ParameterType(initOrder):
    """Base class for all prameter types.

     All available parameter types inherit from this class. The class creates
     a parameter object holding the following user defined information.

     Parameters
     ----------

     name: str
         Name of the parameter.
     default: ParameterBaseType, default: str
         the default value of the given parameter. **Note**: this value must be
         a *valid* parameter value!
    mandatory: bool, default: False
         boolean indicating if this parameter is required
     max_items: int, default: 1
         If set to > 1 it will cause the values to be returned in a
         list (even if the user only provided 1). Raises an error if more than
         than ``max_items`` values are parsed.
     item_separator: str, default: ,
         The string used to separate multiple values for this
         parameter. In some cases (at the shell, web interface, etc) the user have
         always the option to provide multiple values by re-using the same parameter
         name (e.g. ``param1=a param1=b`` produces ``{'param1': ['a', 'b']}``). But the
         configuration file does not allow this at this time. Therefore is better
         to setup a separator, even though the user might not use it while giving
         input. It must not be a character, it can be any string
         (make sure it's not a valid value!!)
     regex: Optional[str], default: None
         A regular expression defining valid "string" values before
         parsing them to their defining classes (e.g. an Integer might define a
         regex of "[0-9]+" to prevent getting negative numbers). This will be used
         also on Javascript so don't use fancy expressions or make sure they are
         understood by both python and Javascript.
     help: str, default: No help available
         The help string describing what this parameter is good for.
     print_format: str, default %s
         String format used to display parameter values, e.g. ``%.2f`` to
         display always 2 decimals for floats
     impact:
         The impact of the parameter to the output, possible values are
         Parameter.Impact.affects_values, Parameter.Impact.affects_plots, Parameter.
         Impact.no_effects


     Properties
     ----------

     base_type:
         Type of this parameter
    """

    _pattern = None  # laizy init.
    base_type: Type[ParameterBaseType] = str
    """Type of this parameter."""

    def __init__(
        self,
        name=None,
        default=None,
        mandatory=False,
        max_items=1,
        item_separator=",",
        regex=None,
        version=1,
        help="No help available.",
        print_format="%s",
        impact=Parameter.Impact.affects_values,
    ):
        """Creates a Parameter with the following information."""
        self.name = name

        self.mandatory = mandatory
        if max_items < 1:
            raise ValidationError(
                "max_items must be set to a value >= 1. Current='%s'" % max_items
            )
        self.max_items = max_items
        self.item_separator = item_separator

        self.regex = regex
        self.help = help

        self.print_format = print_format

        # How important is this setting for configuration?
        self.impact = impact

        # set the version of the field
        self.version = version

        # this assures we get a valid default!
        if default is None:
            self.default = None
        else:
            self.default = self.parse(default)

        self.id = None
        self.is_default = False

    def synchronize(self, tool: str) -> int:
        """Read the id of a tool from the database

        Parameters
        ----------
        tool:
            name of the plugin that is synchronised.

        Returns
        -------
        int:
            database id entry for the plugin
        """
        if self.id is None:
            itype = self.__class__.__name__
            tool_obj = Parameter.objects.get_or_create(
                tool=tool,
                version=self.version,
                mandatory=self.mandatory,
                default=self.default,
                impact=self.impact,
                parameter_name=self.name,
                parameter_type=itype,
            )
            self.id = tool_obj[0].id
        return self.id

    def _verified(
        self,
        orig_values: Any,
    ) -> Any:
        """Check if given values are valid."""

        if not isinstance(orig_values, list):
            values = [orig_values]
        else:
            values = orig_values

        if len(values) > self.max_items:
            raise ValidationError(
                "Expected %s items at most, got %s" % (self.max_items, len(values))
            )

        if (
            self.regex is not None
            and self._pattern is None
            and isinstance(values[0], str)
        ):

            self._pattern = re.compile(self.regex)

        if self._pattern:
            for val in values:
                if isinstance(val, str) and not self._pattern.search(val):
                    raise ValidationError("Invalid Value: %s" % val)

        # so it works transparent
        return orig_values

    def to_str(self, value: Any) -> str:
        """Transform this value in a serializable string.

        Parameters
        ----------
        value: Any
            Value that is to be converted to a string

        Returns
        -------
        str:
            String representation of the input value.
        """
        if self.max_items > 1:
            if not isinstance(value, list):
                value = [value]
            if self.item_separator is not None:
                return self.item_separator.join(value)
            else:
                # assume is a json array:
                return json.dumps(value)

        else:
            return str(value)

    def parse(
        self, value: Union[str, list[str]]
    ) -> Union[list[ParameterBaseType], ParameterBaseType]:
        """Parse a parameter value.

        Parameters
        ----------
        value: Union[str, list[str]]
            The parameter value that should be parsed.

        Returns
        -------
        ParameterBaseType:
            Parsed parameter value
        """
        if self.max_items > 1:
            if isinstance(value, str):
                if self.item_separator is not None:
                    return [
                        self.base_type(v)
                        for v in self._verified(value.split(self.item_separator))
                    ]
                elif value[0] == "[":
                    # assume is a json array:
                    return [
                        self.base_type(v) for v in self._verified(json.loads(value))
                    ]
                else:
                    # this is a single string, but we expect multiple,
                    # so convert appropriately (in list)
                    return [self.base_type(self._verified(value))]
            else:
                # if here assume is iterable, if not see the scp
                try:
                    return [self.base_type(v) for v in self._verified(value)]
                except TypeError:
                    # it was not iterable... but we expect more than one,
                    # so return always a list
                    return [self.base_type(self._verified(value))]

        return self.base_type(self._verified(value))

    def format(self, value: Optional[str] = None) -> str:
        """Format the default value or the given one to a string.

        Parameters
        ----------
        value: Optional[str], default: None
            the value to be formated, if set to None the default value
            of this paramter will be used.

        Returns
        -------
        str:
            formatted string value


        .. note::

            This can be overwriten to provide more control over how values are
            being displayed.
        """
        if value is None:
            if self.default is None:
                return "<undefined>"
            value = self.default
        return self.print_format % value

    def __str__(self):
        return self.__class__.__name__

    @staticmethod
    def infer_type(value: Any) -> Type:
        """Infer the type of a given default."""
        if isinstance(value, type):
            this_type = value
        else:
            this_type = type(value)
        if this_type in _type_mapping:
            return _type_mapping[this_type]()
        raise ValueError("Can't infer type for default '%s'." % value)

    def get_type(self) -> str:
        """Get the name of the class."""

        return self.__class__.__name__


class ParameterDictionary(dict):
    """Directory holding all plugin parameters for a ``Freva`` plugin.

    This class behaves like a built-in ``dict`` with additional features.
    The most prominent feature is that the order of added items is preseverd,
    as opposed to a normal build-in ``dict``.

    Parameters
    ----------
    parameters: :class:`ParameterType`
        collection of parameters of type :class:`ParameterType`.
        Note: The order of the parameters will be preseverd.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.String(name="title", default="The title"),
                parameters.Integer(name="int_value", mandatory=True),
                parameters.Float(name="float_value", default=0.0)
            )

    """

    def __init__(self, *parameters: ParameterType) -> None:
        """Instantiate ParameterDictionary with the given list of parameters."""
        super().__init__()
        extra = config.get_section("scheduler_options").get("extra_options", "").strip()
        if extra.lower() == "none":
            extra = ""
        extra_scheduler_options = String(
            "extra_scheduler_options",
            default=extra,
            help=(
                "Set additional options for the job submission to the "
                "workload manager (, seperated). Note: batchmode and web only."
            ),
        )
        self._params: dict[str, ParameterType] = dict()
        for param in parameters:
            # check name is unique
            if param.name in self._params:
                raise ValueError(
                    "Parameters name must be unique. Got second %s key." % param.name
                )
            self._params[param.name] = param
            self[param.name] = param.default
        self.setdefault("extra_scheduler_options", extra)
        self._params.setdefault("extra_scheduler_options", extra_scheduler_options)

    def __str__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(
                ["%s<%s>: %s" % (k, self._params[k], v) for k, v in self.items()]
            ),
        )

    def get_parameter(self, param_name: str) -> ParameterType:
        """Return the parameter object from the given name.

        Parameters
        ---------

        param_name: str
            Name of the parameter that is queried

        Raises
        ------

        ValidationError:
            if the parameter name doesn't match anything stored here.

        """
        if param_name not in self:
            mesg = "Unknown parameter %s" % param_name
            similar_words = find_similar_words(param_name, self.keys())
            if similar_words:
                mesg = "%s\n Did you mean this?\n\t%s" % (
                    mesg,
                    "\n\t".join(similar_words),
                )
            raise ValidationError(mesg)
        return self._params[param_name]

    def parameters(self) -> list[ParameterType]:
        """All parameters stored in :class:`ParameterDictionary`"""
        return list(self._params.values())

    def _complete(
        self,
        config_dict: Optional[dict[str, Any]] = None,
        add_missing_defaults: bool = False,
    ) -> dict[str, Any]:
        """Completes a given dictionary with default values if required."""
        config_dict = config_dict or {}
        for key in set(self) - set(config_dict):
            if add_missing_defaults or self._params[key].default is not None:
                self._params[key].is_default = True
                default_param = self._params[key].default
                config_dict[key] = default_param
        return config_dict

    def validate_errors(
        self,
        config_dict: dict[str, Any],
        raise_exception: bool = False,
    ) -> dict[str, list[tuple[str, int]]]:
        """Checks if the given configuration dictionary is valied.

        Parameters
        ----------
        config_dict:
            the dictionary validated.
        raise_exception:
            If an exception should be risen. In such a case only a
            message is elevated (this could be changed)

        Returns
        -------
        dict:
            a dictionary with missing items and those having to manay of them or

        Raises
        ------
        ValidationError:
            If parameters are missing/duplicated and a `raise_exception` flag is
            set to `True`.
        """
        missing: list[tuple[str, int]] = []
        too_many_items: list[tuple[str, int]] = []
        for key, param in self._params.items():
            if param.mandatory and config_dict.get(key) is None:
                missing.append((key, param.max_items))
            default_value = config_dict.get(key, [])
            if not isinstance(default_value, list):
                default_value = [default_value]
            if len(default_value) > param.max_items:
                too_many_items.append((key, param.max_items))
        if missing or too_many_items:
            if raise_exception:
                msg = "Error found when parsing parameters. "
                if missing:
                    missing_values = [value for (value, _) in missing]
                    msg += "Missing mandatory parameters: %s" % ", ".join(
                        missing_values
                    )
                if too_many_items:

                    msg += "Too many entries for these parameters: %s" % ", ".join(
                        [
                            "%s(max:%s, found:%s)"
                            % (param, max, len(config_dict[param]))
                            for param, max in too_many_items
                        ]
                    )
                raise ValidationError(msg)
            return dict(missing=missing, too_many_items=too_many_items)
        return {}

    @deprecated_method("ParameterDictionary", "parse_arguments")
    def parseArguments(
        self, *args, **kwargs
    ) -> dict[str, Union[ParameterBaseType, list[ParameterBaseType]]]:
        """Deprecated version of the :class:`parse_arguments` method.

        :meta private:
        """
        return self.parse_arguments(*args, **kwargs)

    def parse_arguments(
        self,
        opt_arr: Union[str, list[str]],
        use_defaults: bool = False,
        complete_defaults: bool = False,
        check_errors: bool = True,
    ) -> dict[str, Union[ParameterBaseType, list[ParameterBaseType]]]:
        """Parse a list of strings to a parameter dictionary.

        The strings are of the type: ``key1=val1`` or ``key2``
        multiple values can be defined by either defining the same key
        multiple times or by using the item_separator character

        Parameters
        ----------
        opt_arr: Union[str, list[str]]
            List of strings containing
            ("key=value"|"key"|"key=value1,value2" iff item_separator==',')
        use_defaults: bool, default: False
            If the parameters defaults should be used when value is missing
        complete_defaults: bool, default: False
            Return a complete configuration containing None for those not
            provided parameters that has no defaults.
        check_errors: bool, default: True
            Check for configuration errors.

        Returns
        -------
        dict[str, Union[ParameterBaseType, list[ParameterBaseType]]:
            dictionary holding the plugin configuration.

        Raises
        ------
        ValidationError:
            Raises a ValidationError if a wrong configuration was parsed.

        """
        param_config: dict[str, Any] = defaultdict(list)
        if not isinstance(opt_arr, (list, tuple, set)):
            opt_arr = [opt_arr]
        for option in opt_arr:
            key, _, value = [key.strip() for key in option.partition("=")]
            value = value or "true"
            try:
                parsed_values: Any = self._params[key].parse(value)
            except KeyError as error:
                raise ValidationError(f"{key} is not a valid parameter") from error
            if isinstance(parsed_values, list):
                if isinstance(param_config[key], list):
                    param_config[key] = param_config[key] + parsed_values
                else:
                    param_config[key] = [param_config[key]] + parsed_values
            elif key not in param_config:
                param_config[key] = parsed_values
            elif not isinstance(parsed_values, list):
                param_config[key] = [param_config[key]] + [parsed_values]
            else:
                param_config[key].append(parsed_values)
        if use_defaults:
            self._complete(param_config, add_missing_defaults=complete_defaults)
        if check_errors:
            self.validate_errors(param_config, raise_exception=True)
        return param_config

    def get_help(self, width: int = 80) -> str:
        """Render plugin help string to be displayed in a cli context.

        Parameters
        ----------
        width: int, default 80
            Column width used to wrap the help text.

        Returns
        -------
        str:
            Help test for this plugin configuration.
        """

        help_str: list[str] = []
        # compute maximal param length for better viewing
        max_size = max([len(k) for k in self] + [0])
        if max_size > 0:
            wrapper = textwrap.TextWrapper(
                width=width,
                initial_indent=" " * (max_size + 1),
                subsequent_indent=" " * (max_size + 1),
                replace_whitespace=False,
            )
            help_str.append("Options:")

            for key, param in self._params.items():
                param_format = "%%-%ss (default: %%s)" % (max_size)
                help_str.append(param_format % (key, param.format()))
                if param.mandatory:
                    help_str[-1] = help_str[-1] + " [mandatory]"

                # wrap it properly
                help_str.append(
                    "\n".join(wrapper.fill(line) for line in param.help.splitlines())
                )
        return "\n".join(help_str)

    def synchronize(self, tool: str) -> None:
        """Synchronize all entries for a plugin configuration of a given tool

        Parameters
        ----------
        tool: str
            Name of the tool to be synced
        """

        for entry in self._params.values():
            entry.synchronize(tool)


class String(ParameterType):
    """A simple string parameter.

    Parameters
    ----------
    kwargs:
        Additional :class:`ParameterType` parameters.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.String(
                            name="plot_tile",
                            mandatory=True,
                            default="Plot title",
                            help="Set a title for the plot"
                            ),
            )


    """

    base_type = str


class Integer(ParameterType):
    """An integer parameter.

    Parameters
    ----------
    kwargs:
        Additional :class:`ParameterType` parameters.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.Integer(
                            name="const",
                            default=3,
                            help="Set a constant factor"
                            ),
            )


    """

    base_type = int

    def __init__(self, **kwargs):
        kwargs["regex"] = r"^[+-]?[0-9]+$"
        super().__init__(**kwargs)


class Float(ParameterType):
    """A float parameter.

    Parameters
    ----------
    kwargs:
        Additional :class:`ParameterType` parameters.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.Float(
                            name="mul",
                            default=3.16,
                            help="Set a multiply factor."
                            ),
            )

    """

    base_type = float

    def __init__(
        self,
        **kwargs,
    ):
        kwargs[
            "regex"
        ] = r"^[+-]?(?:[0-9]+\.?[0-9]*|[0-9]*\.?[0-9]+)(?:[eE][+-]?[0-9]+)?$"
        super().__init__(**kwargs)


class File(String):
    """A parameter representing a file in the system.

    Parameters
    ----------
    file_extension: str, default: nc
        Suffix (file types) of the files that should be considered.
    kwargs:
        Additional :class:`ParameterType` parameters.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.File(
                            name="input_file",
                            file_extension="geojson",
                            mandatory=False,
                            help="Select a geojson file.",
                            ),
            )

    """

    def __init__(self, file_extension="nc", **kwargs):
        self.file_extension = file_extension
        super().__init__(**kwargs)


class Directory(String):
    """A parameter representing a directory in the system. Deprecated

    :meta private:
    """

    def __init__(self, impact=Parameter.Impact.no_effects, **kwargs):
        super().__init__(impact=impact, **kwargs)


class InputDirectory(String):
    """A parameter representing a input directory in the system.

    Parameters
    ----------
    kwargs:
        Additional :class:`ParameterType` parameters.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.InputDirectory(
                            name="input_directory",
                            default="/work/data",
                            mandatory=False,
                            help="Select the input directory",
                            ),
            )

    """


class CacheDirectory(Directory):
    """A parameter representing a cache directory in the system.

    Parameters
    ----------
    kwargs:
        Additional :class:`ParameterType` parameters.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.CacheDirectory(
                            name="cache_directory",
                            default="/scratch",
                            help="Set the path to a temporary directory",
                            ),
            )

    """


class Date(String):
    """A date parameter.

    Parameters
    ----------
    kwargs:
        Additional :class:`ParameterType` parameters.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.Date(
                            name="time",
                            default="1950-01-10",
                            help="Select a timestamp",
                            ),
            )

    """

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)


class Bool(ParameterType):
    """A boolean parameter.

    Parameters
    ----------

    kwargs:
        Additional :class:`ParameterType` parameters.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.Bool(
                            name="convert_variables",
                            default=True,
                            help="Convert the variables",
                            ),
            )

    """

    base_type = bool

    def parse(self, value: Any) -> bool:
        """Convert a string to a boolean value.

        The following values will be mapped

          true, t, yes, y, on, 1 => TRUE
          false, f, no, n, off, 2 => FALSE

        Parameters
        ----------

        value: Union[str, int, bool]
            Input representation of the boolean such as
            true, false, t, y, f, ...


        Raises
        ------
        ValidationException:
            if the given input value can not be converted to a bool.
        """

        if isinstance(value, bool):
            # it was no bool after all...
            return value
        if isinstance(value, str) and value:
            if value.lower() in ["true", "t", "yes", "y", "on", "1"]:
                return True
            if value.lower() in ["false", "f", "no", "n", "off", "0"]:
                return False
        elif isinstance(value, int):
            return bool(value)
        # if here we couldn't parse it
        raise ValueError(f"'{value}' is can't be converted to a boolean")


_type_mapping = {str: String, int: Integer, float: Float, bool: Bool}
"""These are the mapping of the default Python types to those of this framework.
This is mapping is used by the infer_type function to infer the type of a
given parameter default."""


class Range(String):
    """
    A range parameter, e.g passing experiment lists (1970,1975,...,2000).

    Parameters
    ---------

    see :class:`ParameterType` parameters for more details.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.Range(
                            name="time_range",
                            default="1950:10:1980", # range(1950, 1990, 10)
                            help="Set a time range",
                            ),
            )

    """

    def __init__(self, *args, **kwargs):
        # set max_items to a very large number...
        kwargs["max_items"] = 1e20
        self.base_type = int  # self.PrintableList
        super().__init__(*args, **kwargs)

    def _parse_colon(self, string: str) -> list[int]:
        """
        Parse colon separated strings.

        """
        tmp = list(map(int, string.split(":")))
        if len(tmp) == 2:
            return list(range(tmp[0], tmp[1] + 1))
        elif len(tmp) == 3:
            return list(range(tmp[0], tmp[2] + 1, tmp[1]))
        elif len(tmp) == 1:
            return tmp
        else:
            raise ValueError("'%s' is no recognized as a range value" % str)

    def _parse_comma(self, string: str) -> list[int]:
        """
        Parses comma separated strings and checks each part for colon
        separated strings

        Parameters
        ----------

        string: str
            comma seperated values that are converted to list

        Returns
        -------

        list[int]:
            list with integers
        """
        parts = string.split(",")
        res_list = list()
        for part in parts:
            res_list += self._parse_colon(part)
        return res_list

    def parse(self, value: Any) -> PrintableList:
        """
        Parses a "RangeString" and returns a "PrintableList".

        Values can be a comma separated string, colon separated or a combination
        Values after "-" are deleted from the resulting list

        Parameters
        ----------
        value: 'start:step:stop-value' --> 1970:5:2000-1985 or 1970:2000,1980-1990:1995


        Returns
        -------
        PrintableList:
            PrintableList of range items

        Raises
        ------
        ValueError:
            Raises a ValueError if items can't be parsed to list.
        """
        try:
            main_parts = value.split("-")
            result = self._parse_comma(main_parts[0])
            del_list = PrintableList()
            for part in main_parts[1:]:
                del_list += self._parse_comma(part)
            return PrintableList(sorted([x for x in result if x not in del_list]))
        except AttributeError as err:
            raise ValueError(f"'{value}' is no recognized as a range value") from err

    def to_str(self, value: Any) -> str:
        """Conevert input value to string."""
        return str(value)


class Unknown(String):
    """An unknown parameter for conversions.

    Parameters
    ---------
    see :class:`ParameterType` parameters for more details.
    """

    def __init__(
        self, impact=Parameter.Impact.affects_values, mandatory=True, **kwargs
    ):
        super().__init__(impact=impact, mandatory=mandatory, **kwargs)


class SolrField(String):
    """
    A parameter using solr for finding valid values.

    Parameters
    ---------
    facet: str
        Solr search facet used for this parameter
    group: int, default: 1
        The group this search facet belongs to. This can be used to group
        different search facets together, for example for comparing
        multi model ensemble
    multiple: bool, default: False
        flag indicating whether multiple facets are allowed.
    predefined_facets: Optional[list[str]], default: None
        a list of strings that are set as default search facets
    editable:
        flag indicating whether or not the value can be changed.
    kwargs:
        additional :class:`ParameterType` parameters.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.SolrField(
                                name="variable_name",
                                default="tas",
                                facet="variable",
                                max_items=1,
                                group=2,
                                predefined_facets={"time_frequency":["1hr"]},
                                help="Select the variable name",
                            ),
            )

    """

    def __init__(
        self,
        *args,
        facet: Optional[str] = None,
        group: int = 1,
        multiple: bool = False,
        predefined_facets: Optional[list[str]] = None,
        editable: bool = True,
        **kwargs,
    ):
        if not facet:
            raise TypeError("`facet` must not be empty")
        self.facet = facet
        self.group = group
        self.multiple = multiple
        self.predefined_facets = predefined_facets
        self.editable = editable
        super().__init__(*args, **kwargs)


class SelectField(String):
    """Select field to select parameter from predefined vlaues.

    Parameters
    ----------
    options: dict[str, str]
        Directory representing the names and values of the predefined options.
    kwargs:
        Additional :class:`ParameterType` parameters.

    Example
    -------

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __parameters__ = parameters.ParameterDictionary(
                parameters.SelectField(
                                    name="options",
                                    default="first",
                                    options={1: "first", 2: "second"}
                                    help="Select from options",
                                      ),
            )

    """

    def __init__(self, options: dict[str, str], *args, **kwargs):
        self.options = options
        super().__init__(*args, **kwargs)

    def _verified(self, orig_values):

        if orig_values not in self.options.values():
            values = ",".join(list(self.options.values()))
            raise ValueError(
                f'Only the following values are allowed for "{self.name}": {values}'
            )
        return orig_values

    def parse(self, value: Any) -> str:
        """Parse a parameter value.

        Parameters
        ----------
        value: Union[str, list[str]]
            The parameter value that should be parsed.

        Returns
        -------
        ParameterBaseType:
            Parsed paramter value

        Raises
        ------
        ValueError:
            if value is not part of possible options.
        """

        if self._verified(value):
            for key, val in self.options.items():
                if value == val:
                    return key
        return ""



import pytest

from evaluation_system.api.workload_manager.moab import MoabJob
from evaluation_system.api.workload_manager.pbs import PBSJob


@pytest.mark.parametrize("Cluster", [PBSJob, MoabJob])
def test_header(Cluster):
    with Cluster(
        walltime="00:02:00", processes=4, cores=8, memory="28GB", name="worker"
    ) as cluster:

        assert "#PBS" in cluster.job_header
        assert "#PBS -N worker" in cluster.job_header
        assert "#PBS -l select=1:ncpus=8:mem=27GB" in cluster.job_header
        assert "#PBS -l walltime=00:02:00" in cluster.job_header
        assert "#PBS -q" not in cluster.job_header
        assert "#PBS -A" not in cluster.job_header

    with Cluster(
        queue="regular",
        project="DaskOnPBS",
        processes=4,
        cores=8,
        resource_spec="select=1:ncpus=24:mem=100GB",
        memory="28GB",
    ) as cluster:

        assert "#PBS -q regular" in cluster.job_header
        assert "#PBS -N worker" in cluster.job_header
        assert "#PBS -l select=1:ncpus=24:mem=100GB" in cluster.job_header
        assert "#PBS -l select=1:ncpus=8:mem=27GB" not in cluster.job_header
        assert "#PBS -l walltime=" in cluster.job_header
        assert "#PBS -A DaskOnPBS" in cluster.job_header

    with Cluster(cores=4, memory="8GB") as cluster:

        assert "#PBS -j oe" not in cluster.job_header
        assert "#PBS -N" in cluster.job_header
        assert "#PBS -l walltime=" in cluster.job_header
        assert "#PBS -A" not in cluster.job_header
        assert "#PBS -q" not in cluster.job_header

    with Cluster(cores=4, memory="8GB", job_extra=["-j oe"]) as cluster:

        assert "#PBS -j oe" in cluster.job_header
        assert "#PBS -N" in cluster.job_header
        assert "#PBS -l walltime=" in cluster.job_header
        assert "#PBS -A" not in cluster.job_header
        assert "#PBS -q" not in cluster.job_header


@pytest.mark.parametrize("Cluster", [PBSJob, MoabJob])
def test_job_script(Cluster):
    with Cluster(walltime="00:02:00", processes=4, cores=8, memory="28GB") as cluster:

        job_script = cluster.job_script()
        assert "#PBS" in job_script
        assert "#PBS -N worker" in job_script
        assert "#PBS -l select=1:ncpus=8:mem=27GB" in job_script
        assert "#PBS -l walltime=00:02:00" in job_script
        assert "#PBS -q" not in job_script
        assert "#PBS -A" not in job_script

    with Cluster(
        queue="regular",
        project="DaskOnPBS",
        processes=4,
        cores=8,
        resource_spec="select=1:ncpus=24:mem=100GB",
        memory="28GB",
    ) as cluster:

        job_script = cluster.job_script()
        assert "#PBS -q regular" in job_script
        assert "#PBS -N worker" in job_script
        assert "#PBS -l select=1:ncpus=24:mem=100GB" in job_script
        assert "#PBS -l select=1:ncpus=8:mem=27GB" not in job_script
        assert "#PBS -l walltime=" in job_script
        assert "#PBS -A DaskOnPBS" in job_script


def test_informative_errors():
    with pytest.raises(ValueError) as info:
        PBSJob(memory=None, cores=4)
    assert "memory" in str(info.value)



from collections import namedtuple
from configparser import ConfigParser, ExtendedInterpolation
from datetime import datetime
from getpass import getuser
import os
from pathlib import Path
import shutil
import socket
from tempfile import TemporaryDirectory, NamedTemporaryFile
import time

import django
from django.conf import settings
import pytest
import toml
import mock

from evaluation_system.misc import config


def get_config(admin=False):
    from .mocks import TEST_EVAL, TEST_DRS

    test_cfg = ConfigParser(interpolation=ExtendedInterpolation())
    test_cfg.read_string(TEST_EVAL)
    drs_cfg = toml.loads(TEST_DRS)
    cfg_p = ConfigParser(interpolation=ExtendedInterpolation())
    if admin:
        test_cfg["evaluation_system"].setdefault("admins", getuser())
    items_to_overwrite = [
        "db.host",
        "db.user",
        "db.passwd",
        "db.db",
        "db.port",
        "solr.host",
        "solr.port",
        "solr.core",
    ]
    cfg_p.read(os.environ["EVALUATION_SYSTEM_CONFIG_FILE"])
    cfg = dict(cfg_p["evaluation_system"].items())
    for key in items_to_overwrite:
        value = test_cfg["evaluation_system"].get(key)
        test_cfg.set("evaluation_system", key, cfg.get(key, value))
    return test_cfg, drs_cfg


def mock_config(keyfile, admin=False, patch_env=True):
    cfg, drs_config = get_config(admin)
    with TemporaryDirectory() as temp_dir:
        eval_config = Path(temp_dir) / "evaluation_system.conf"
        drs_conf_path = Path(temp_dir) / "drs_config.toml"
        crawl_data_dir = Path(temp_dir) / "data" / "crawl_my_data"
        drs_config["crawl_my_data"]["root_dir"] = str(crawl_data_dir)
        PATH = (Path(__file__).parent / "mocks" / "bin").absolute()
        env = dict(
            EVALUATION_SYSTEM_CONFIG_FILE=str(eval_config),
            EVALUATION_SYSTEM_DRS_CONFIG_FILE=str(drs_conf_path),
            PUBKEY=str(keyfile),
            PATH=str(PATH) + ":" + os.environ["PATH"],
        )
        with open(eval_config, "w") as f:
            cfg.write(f)
        with open(drs_conf_path, "w") as f:
            toml.dump(drs_config, f)
        if not patch_env:
            yield env
        else:
            with mock.patch.dict(os.environ, env, clear=True):
                yield config
    try:
        shutil.rmtree(config.get("base_dir_location"))
    except:
        pass


@pytest.fixture(scope="function")
def temp_script():

    with NamedTemporaryFile(suffix="test.sh") as tf:
        yield tf.name


@pytest.fixture(scope="session")
def dummy_key():
    with NamedTemporaryFile(suffix=".crt") as tf:
        with Path(tf.name).open("w") as f:
            f.write("------ PUBLIC KEY ----\n12345\n---- END PUBLIC KEY ----")
        yield tf.name


@pytest.fixture(scope="module")
def plugin_doc():

    with TemporaryDirectory() as td:
        dummy_doc = Path(td) / "dummy_plugin_doc.tex"
        dummy_bib = Path(td) / "dummy_plugin.bib"
        with (dummy_doc).open("w") as f:
            f.write(
                """\\documentclass[12pt]{article}
\\usepackage[utf8]{inputenc}
\\begin{document}
This is a dummy doc
\\end{document}"""
            )
        with (dummy_doc).open("w") as f:
            f.write(
                """% This file was created with JabRef 2.9.2.
% Encoding: UTF-8"""
            )
        yield dummy_doc


@pytest.fixture(scope="session")
def admin_env(dummy_key):
    yield from mock_config(dummy_key, admin=True, patch_env=False)


@pytest.fixture(scope="session")
def dummy_env(dummy_key):
    yield from mock_config(dummy_key, admin=False)


@pytest.fixture(scope="session")
def git_config():
    yield "git config init.defaultBranch main; git config user.name your_user; git; config user.email your@email.com"


@pytest.fixture(scope="module")
def dummy_git_path():
    with TemporaryDirectory(prefix="git") as td:
        repo_path = Path(td) / "test_plugin.git"
        tool_path = Path(td) / "test_tool"
        tool_path.mkdir(exist_ok=True, parents=True)
        repo_path.mkdir(exist_ok=True, parents=True)
        yield repo_path, tool_path


@pytest.fixture(scope="module")
def temp_dir():
    with TemporaryDirectory() as td:
        yield Path(td)


@pytest.fixture(scope="function")
def dummy_crawl(dummy_solr, dummy_settings, dummy_env):

    # At this point files have been ingested into the server already,
    # delete them again
    [getattr(dummy_solr, key).delete("*") for key in ("all_files", "latest")]
    from evaluation_system.misc import config

    root_path = Path(config.get_drs_config()["crawl_my_data"]["root_dir"])
    crawl_dir = root_path / f"user-{getuser()}"
    user_files = []
    for file in map(Path, dummy_solr.files):
        # Drop the version from the files
        file_parts = file.parts[1:-2] + (file.parts[-1],)
        crawl_file = crawl_dir.joinpath(*file_parts)
        crawl_file.parent.mkdir(exist_ok=True, parents=True)
        crawl_file.touch()
        user_files.append(crawl_file)
    yield user_files
    [getattr(dummy_solr, key).delete("*") for key in ("all_files", "latest")]
    shutil.rmtree(root_path)


@pytest.fixture(scope="module")
def dummy_solr(dummy_env, dummy_settings):

    dummy_settings.reloadConfiguration()
    server = namedtuple(
        "solr",
        [
            "solr_port",
            "solr_host",
            "all_files",
            "latest",
            "tmpdir",
            "drsfile",
            "files",
            "cmd",
            "DRSFILE",
        ],
    )
    server.solr_port = dummy_settings.get("solr.port")
    server.solr_host = dummy_settings.get("solr.host")
    from evaluation_system.model.solr_core import SolrCore
    from evaluation_system.model.solr import SolrFindFiles
    from evaluation_system.model.file import DRSFile
    from evaluation_system.misc.utils import supermakedirs

    server.all_files = SolrCore(
        core="files", host=server.solr_host, port=server.solr_port
    )
    server.latest = SolrCore(
        core="latest", host=server.solr_host, port=server.solr_port
    )
    server.all_files.delete("*")
    server.latest.delete("*")
    with TemporaryDirectory(prefix="solr") as td:
        supermakedirs(str(Path(td) / "solr_core"), 0o0777)
        server.tmpdir = str(Path(td) / "solr_core")
        DRSFile._load_structure_definitions()
        orig_dir = DRSFile.DRS_STRUCTURE["cmip5"].root_dir
        DRSFile.DRS_STRUCTURE["cmip5"].root_dir = server.tmpdir
        DRSFile.DRS_STRUCTURE_PATH_TYPE[server.tmpdir] = "cmip5"
        server.files = [
            "cmip5/output1/MOHC/HadCM3/historical/mon/aerosol/aero/r2i1p1/v20110728/wetso2/wetso2_aero_HadCM3_historical_r2i1p1_190912-193411.nc",
            "cmip5/output1/MOHC/HadCM3/decadal2008/mon/atmos/Amon/r9i3p1/v20120523/tauu/tauu_Amon_HadCM3_decadal2008_r9i3p1_200811-201812.nc",
            "cmip5/output1/MOHC/HadCM3/decadal2009/mon/atmos/Amon/r7i2p1/v20110719/ua/ua_Amon_HadCM3_decadal2009_r7i2p1_200911-201912.nc",
            "cmip5/output1/MOHC/HadCM3/decadal2009/mon/atmos/Amon/r7i2p1/v20110819/ua/ua_Amon_HadCM3_decadal2009_r7i2p1_200911-201912.nc",
            "cmip5/output1/MOHC/HadCM3/decadal2009/mon/atmos/Amon/r7i2p1/v20110419/ua/ua_Amon_HadCM3_decadal2009_r7i2p1_200911-201912.nc",
        ]
        for f in server.files:
            abs_path = Path(server.tmpdir) / f
            abs_path.parent.mkdir(exist_ok=True, parents=True)
            with abs_path.open("w") as f_out:
                f_out.write(" ")
        data_dir = Path(server.tmpdir) / "cmip5"
        # add the files to solr
        SolrCore.load_fs(
            data_dir,
            abort_on_errors=True,
            core_all_files=server.all_files,
            core_latest=server.latest,
        )
        server.DRSFile = DRSFile
        server.fn = str(Path(server.tmpdir) / server.files[0])
        server.drs = server.DRSFile.from_path(server.fn)
        yield server
    server.all_files.delete("*")
    server.latest.delete("*")
    DRSFile.DRS_STRUCTURE["cmip5"].root_dir = orig_dir


@pytest.fixture(scope="module")
def django_user(dummy_settings, dummy_env):

    from django.contrib.auth.models import User
    from evaluation_system.model.history.models import History
    from django.contrib.auth.models import User

    user_django, created = User.objects.get_or_create(username=getuser())
    yield user_django
    User.objects.filter(username=getuser()).delete()
    History.objects.all().delete()


@pytest.fixture(scope="module")
def dummy_config(dummy_env, dummy_settings_single):

    from evaluation_system.misc import config, utils

    config.reloadConfiguration()
    yield config
    config.reloadConfiguration()


@pytest.fixture(scope="module")
def dummy_plugin(dummy_env, dummy_settings):

    from evaluation_system.tests.mocks.dummy import DummyPlugin

    yield DummyPlugin()


@pytest.fixture(scope="function")
def dummy_history(dummy_env, dummy_settings):

    from evaluation_system.model.history.models import History

    yield History
    History.objects.all().delete()


@pytest.fixture(scope="module")
def test_user(dummy_env, dummy_settings, config_dict):

    from evaluation_system.model.history.models import History
    from django.contrib.auth.models import User

    user = User.objects.create_user(username="test_user2", password="123")
    hist = History.objects.create(
        timestamp=datetime.now(),
        status=History.processStatus.running,
        uid=user,
        configuration='{"some": "config", "dict": "values"}',
        tool="dummytool",
        slurm_output="/path/to/slurm-44742.out",
        host=socket.gethostbyname(socket.gethostname()),
    )
    yield user, hist
    user.delete()
    hist


@pytest.fixture(scope="function")
def temp_user(dummy_settings):
    from evaluation_system.tests.mocks.dummy import DummyUser
    import evaluation_system.api.plugin_manager as pm

    with DummyUser(random_home=True, pw_name=getuser()) as user:
        yield user


@pytest.fixture(scope="function")
def dummy_user(
    dummy_env, dummy_settings, config_dict, dummy_plugin, dummy_history, temp_user
):

    from django.contrib.auth.models import User

    User.objects.filter(username="test_user2").delete()
    user_entry = namedtuple("test_user2", ["user", "row_id", "username"])
    user_entry.user = temp_user
    user_entry.username = "test_user2"
    user_entry.row_id = temp_user.getUserDB().storeHistory(
        dummy_plugin,
        config_dict,
        user_entry.username,
        dummy_history.processStatus.not_scheduled,
        caption="My caption",
    )
    yield user_entry
    User.objects.filter(username=user_entry.username).delete()


@pytest.fixture(scope="module")
def config_dict():
    yield {
        "the_number": 42,
        "number": 12,
        "something": "else",
        "other": "value",
        "input": "/folder",
        "variable": "pr",
    }


@pytest.fixture(scope="module")
def tmp_dir():
    with TemporaryDirectory(prefix="freva_test_") as td:
        yield Path(td)
        [f.unlink() for f in Path(td).rglob("*.*")]


@pytest.fixture(scope="module")
def search_dict():
    yield {
        "variable": "tas",
        "project": "CMIP5",
        "product": "output1",
        "time_frequency": "mon",
        "experiment": "decadal2000",
        "model": "MPI-ESM-LR",
        "ensemble": "r1i1p1",
    }


@pytest.fixture(scope="module")
def dummy_settings_single(dummy_env):
    # Application definition
    import evaluation_system.settings.database
    from evaluation_system.misc import config

    config.reloadConfiguration()
    yield config


@pytest.fixture(scope="session")
def dummy_settings(dummy_env):
    from evaluation_system.misc import config

    config.reloadConfiguration()
    import evaluation_system.settings.database

    yield config


@pytest.fixture(scope="function")
def root_path_with_empty_config(dummy_env):
    from evaluation_system.misc import config

    root_path = Path(config.get_drs_config()["crawl_my_data"]["root_dir"])
    crawl_dir = root_path / f"user-{getuser()}" / "freva-ces-plugin-results"
    config._config = {}
    yield crawl_dir
    config.reloadConfiguration()


@pytest.fixture(scope="module")
def hist_obj(django_user):

    from evaluation_system.model.history.models import History
    from django.contrib.auth.models import User
    from evaluation_system.misc import config

    yield History.objects.create(
        status=History.processStatus.running,
        slurm_output="/some/out.txt",
        host=socket.gethostbyname(socket.gethostname()),
        timestamp=datetime.now(),
        uid=User.objects.first(),
    )
    config.reloadConfiguration()



"""The modules provides interfaces for userser to interact with 
the freva data stack."""

from __future__ import annotations
import os
from pathlib import Path
from typing import Any, Generator, Iterator, Union

import lazy_import
from evaluation_system.misc import config
from evaluation_system.misc.exceptions import ConfigurationException

xr = lazy_import.lazy_module("xarray")


class DataReader:
    """Read meta data facets from a collection of datafiles.

    Parameters
    ----------
    paths: os.PathLike
        Input file or input directory of files containing the metadata.
    **defaults: str
        Any default facets that should be assigned.
    """

    suffixes: tuple[str, ...] = (".nc", ".nc4", ".grb", ".grib", ".zarr")
    """Allowed filetypes for files holding meta data."""

    file_sep: str = "_"
    """Character that separates facet values in the file name."""

    drs_specification: str = "crawl_my_data"
    """The drs holding metadata for user data information."""

    def __init__(self, paths: os.PathLike, **defaults: str) -> None:

        self.paths = Path(paths)
        self.defaults = defaults
        drs_config: dict[str, Any] = config.get_drs_config()[self.drs_specification]
        self.root_dir = Path(drs_config["root_dir"]).expanduser().absolute()
        self.parts_dir: list[str] = [
            d for d in drs_config["parts_dir"] if d != "file_name"
        ]
        self.parts_file: list[str] = drs_config["parts_file_name"]

    @staticmethod
    def get_output_directory() -> Path:
        """Get the user data output directory."""
        root_dir = (
            Path(config.get_drs_config()[DataReader.drs_specification]["root_dir"])
            .expanduser()
            .absolute()
        )
        return root_dir

    def __iter__(self) -> Generator[Path, None, None]:
        """Iterate over all found data files."""
        file_iter: Union[Iterator[Path], list[Path]] = []
        if self.paths.is_file():
            file_iter = [self.paths]
        elif self.paths.is_dir():
            file_iter = self.paths.rglob("*")
        else:
            # This is a shot into the dark assumes that the paths variable
            # is a glob pattern
            file_iter = self.paths.parent.rglob(self.paths.name)
        for file in file_iter:
            if file.suffix in self.suffixes:
                yield file

    @property
    def time_table(self) -> dict[int, str]:
        """Tranlatetion from seconds to cmor frequency."""
        return {
            315360000: "dec",
            31104000: "yr",
            2538000: "mon",
            1296000: "sem",
            84600: "day",
            21600: "6h",
            10800: "3h",
            3600: "hr",
            1: "subhr",
        }

    def _timedelta_to_cmor_frequency(self, dt: float) -> str:
        for total_seconds, frequency in self.time_table.items():
            if dt >= total_seconds:
                return frequency
        return "fx"

    def get_time_frequency(self, time_delta: int, freq_attr: str = "") -> str:
        """Create a comor time frequency facet.

        Parameters
        ----------
        time_delta: int
            time delta, in seconds, between consecutive time stpes.
        freq_attr: str, default: ""
            cmor time_frequency attribute that might already be in the data.

        Returns
        -------
        str:
            Matching cmor time frequency.
        """
        if freq_attr in list(self.time_table.values()):
            return freq_attr
        return self._timedelta_to_cmor_frequency(time_delta)

    def get_metadata(self, file_name: os.PathLike) -> dict[str, str]:
        """Read the metadata information from a given file.

        Parameters
        ----------
        file_name: os.PathLike
            The input file the meta data is read from.

        Returns
        -------
        dict[str, str]:
            Dictionary holding meta data information as key-value pair.

        Raises
        ------
        ValueError: If data could not be retrieved.
        """

        try:
            with xr.open_mfdataset(
                str(file_name), parallel=True, use_cftime=True
            ) as dset:
                time_freq = dset.attrs.get("frequency", "")
                data_vars = map(str, dset.data_vars)
                coords = map(str, dset.coords)
                try:
                    times = dset["time"].values[:]
                except (KeyError, IndexError, TypeError):
                    times = []
        except Exception as error:
            raise ValueError(
                f"Could not open data file {file_name}: {error}"
            ) from error
        if len(times) > 0:
            time_str = "-".join(
                [t.strftime("%Y%m%d%H%M") for t in (times[0], times[-1])]
            )
        else:
            time_str = "fx"
        if len(times) > 1:
            dt = abs((times[1] - times[0]).total_seconds())
        else:
            dt = 0
        variables: list[str] = []
        for var in data_vars:
            if var in coords:
                continue
            if (
                "lon" in var.lower()
                or "bnds" in var.lower()
                or "lat" in var.lower()
                or var.lower() in ("x", "y")
            ):
                # Usually those varaialbes should be flagged as coordinates
                # but we can't guaranty that they are.
                continue
            if var.lower() in ["rotated_pole", "rot_pole"]:
                # Also don't consider rotated pole variables
                continue
            variables.append(var)
        if len(variables) != 1:
            raise ValueError(f"Only one data variable allowed found: {variables}")
        _data = self.defaults.copy()
        _data.setdefault("variable", variables[0])
        _data.setdefault("time_frequency", self.get_time_frequency(dt, time_freq))
        _data["time"] = time_str
        return _data

    def file_name_from_metdata(self, path: os.PathLike) -> Path:
        """Construct file name matching the DRS Spec. from given input path.

        Parameters
        ----------
        path:
            Input file holing the meta data information.

        Returns
        -------
        os.PathLike:
            A file name matching the DRS sepcifications so that it can be
            ingested into the solr server.

        Raises
        ------
        ValueError:
            If data could not all metadata could be retrieved.
        """
        path = Path(path)
        meta_data = self.get_metadata(path)
        try:
            dir_path = self.root_dir.joinpath(*[meta_data[d] for d in self.parts_dir])
            file_path = self.file_sep.join([meta_data[d] for d in self.parts_file])
        except KeyError as error:
            raise ValueError(
                f"Not all information could be retrieved: {error}"
            ) from error

        out_dir = (dir_path / file_path).with_suffix(path.suffix)
        return out_dir



from __future__ import annotations
import argparse
from pathlib import Path
import sys
from typing import Any, Optional


from evaluation_system import __version__
from evaluation_system.misc.exceptions import (
    PluginNotFoundError,
    ParameterNotFoundError,
    ValidationError,
    hide_exception,
)
from evaluation_system.misc import logger
import freva
from .utils import BaseCompleter, BaseParser

CLI = "PluginCli"


class PluginCli(BaseParser):
    """Class that constructs the Plugin Argument Parser."""

    desc = "Apply data analysis plugin."

    def __init__(
        self,
        command: str = "freva",
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the plugin sub arg. parser."""
        subparser = parser or argparse.ArgumentParser(
            prog=f"{command}-plugin",
            description=self.desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        subparser.add_argument(
            "tool-name",
            nargs="?",
            metavar="plugin_name",
            help="Plugin name",
            default=None,
        )
        subparser.add_argument(
            "--repo-version",
            default=False,
            action="store_true",
            help="Show the version number from the repository",
        )
        subparser.add_argument(
            "--caption",
            default="",
            help="Set a caption for the results",
        )
        subparser.add_argument(
            "--save",
            default=False,
            action="store_true",
            help="Save the plugin configuration to default destination.",
        )
        subparser.add_argument(
            "--save-config",
            type=Path,
            default=None,
            help="Save the plugin configuration.",
        )
        subparser.add_argument(
            "--show-config",
            help="Show the resulting configuration (implies dry-run).",
            action="store_true",
            default=False,
        )
        subparser.add_argument(
            "--scheduled-id",
            default=None,
            type=int,
            help=argparse.SUPPRESS,
        )
        subparser.add_argument(
            "--dry-run",
            default=False,
            action="store_true",
            help="Perform no computation. Useful for development.",
        )
        subparser.add_argument(
            "--batchmode",
            help="Create a Batch job and submit it to the scheduling system.",
            default=False,
            action="store_true",
        )
        subparser.add_argument(
            "--unique_output",
            help="Append a freva run id to every output folder",
            default=True,
            type=bool,
        )
        subparser.add_argument(
            "--pull-request",
            help="Issue a new pull request for the tool",
            default=False,
            action="store_true",
        )
        subparser.add_argument(
            "--tag", help="Use git commit hash", type=str, default=None
        )
        subparser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        subparser.add_argument(
            "--list-tools",
            "--list",
            "-l",
            default=False,
            action="store_true",
            help="Only list the available tools.",
        )
        subparser.add_argument(
            "--doc",
            "--plugin-doc",
            default=False,
            action="store_true",
            help="Display plugin documentation",
        )
        self.parser = subparser
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(
        args: argparse.Namespace, other_args: Optional[list[str]] = None, **kwargs: Any
    ) -> None:
        """Call the databrowser command and print the results."""
        tool_name = kwargs.pop("tool-name")
        if kwargs.pop("list_tools") or not tool_name:
            print(freva.get_tools_list())
            return
        options: dict[str, Any] = BaseCompleter.arg_to_dict(other_args or [])
        for key, val in options.items():
            if len(val) == 1:
                options[key] = val[0]
        tool_args = {**kwargs, **options}
        try:
            if tool_args.pop("doc"):
                print(freva.plugin_doc(tool_name))
                return
            value, out = freva.run_plugin(tool_name or "", **tool_args)
        except (PluginNotFoundError, ValidationError, ParameterNotFoundError) as e:
            if args.debug:
                raise e
            with hide_exception():
                raise e
        if value != 0:
            logger.warning("Tool failed to run")
        if out:
            print(out)


def main(argv: Optional[list[str]] = None) -> None:
    """Wrapper for entry point script."""
    cli = PluginCli("freva")
    cli.parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    args = cli.parse_args(argv or sys.argv[1:])
    try:
        cli.run_cmd(args, **cli.kwargs)
    except KeyboardInterrupt:  # pragma: no cover
        print("KeyboardInterrupt, exiting", file=sys.stderr, flush=True)
        sys.exit(130)



"""Submit jobs using the oar workload manager."""
from __future__ import annotations
import logging
import shlex
from typing import Optional, Union, ClassVar

from .core import Job

logger = logging.getLogger(__name__)


class OARJob(Job):

    # Override class variables
    submit_command: ClassVar[str] = "oarsub"
    cancel_command: ClassVar[str] = "oardel"
    job_id_regexp: ClassVar[str] = r"OAR_JOB_ID=(?P<job_id>\d+)"
    config_name: ClassVar[str] = "oar"

    def __init__(
        self,
        scheduler: Optional[str] = None,
        name: Optional[str] = None,
        queue: Optional[str] = None,
        project: Optional[str] = None,
        resource_spec: Optional[str] = None,
        walltime: str = "",
        job_extra: Optional[list[str]] = None,
        **base_class_kwargs,
    ):
        super().__init__(scheduler=scheduler, name=name, **base_class_kwargs)
        job_extra = job_extra or []
        header_lines = []
        if self.job_name is not None:
            header_lines.append("#OAR -n %s" % self.job_name)
        if queue is not None:
            header_lines.append("#OAR -q %s" % queue)
        if project is not None:
            header_lines.append("#OAR --project %s" % project)

        # OAR needs to have the resource on a single line otherwise it is
        # considered as a "moldable job" (i.e. the scheduler can chose between
        # multiple sets of resources constraints)
        resource_spec_list = []
        if resource_spec is None:
            # default resource_spec if not specified. Crucial to specify
            # nodes=1 to make sure the cores allocated are on the same node.
            resource_spec = "/nodes=1/core=%d" % self.worker_cores
        resource_spec_list.append(resource_spec)
        if walltime is not None:
            resource_spec_list.append("walltime=%s" % walltime)

        full_resource_spec = ",".join(resource_spec_list)
        header_lines.append("#OAR -l %s" % full_resource_spec)
        header_lines.extend(["#OAR %s" % arg for arg in job_extra])

        self.job_header = "\n".join(header_lines)

        logger.debug("Job script: \n %s" % self.job_script())

    async def _submit_job(self, fn):
        # OAR specificity: the submission script needs to exist on the worker
        # when the job starts on the worker. This is different from other
        # schedulers that only need the script on the submission node at
        # submission time. That means that we can not use the same strategy as
        # in JobQueueCluster: create a temporary submission file, submit the
        # script, delete the submission file. In order to reuse the code in
        # the base JobQueueCluster class, we read from the temporary file and
        # reconstruct the command line where the script is passed in as a
        # string (inline script in OAR jargon) rather than as a filename.
        with open(fn) as f:
            content_lines = f.readlines()

        oar_lines = [line for line in content_lines if line.startswith("#OAR ")]
        oarsub_options = [line.replace("#OAR ", "").strip() for line in oar_lines]
        inline_script_lines = [
            line for line in content_lines if not line.startswith("#")
        ]
        inline_script = "".join(inline_script_lines)
        oarsub_command = " ".join([self.submit_command] + oarsub_options)
        oarsub_command_split = shlex.split(oarsub_command) + [inline_script]
        return self._call(oarsub_command_split)



"""
Created on 17.04.2012

@author: Robert Schuster
"""

import subprocess


class Logger(object):
    """
    The logger class provides static methodes to print colored text.
    """

    # the info should not be colored by default
    color_info = None
    color_error = "red"
    cols = None

    @classmethod
    def Error(cls, text, errorcode=0):
        """
        Print the text in red on the screen with the prefix ERROR:"
        Set the error code to something different from zero to terminate the program
        """
        print(cls.colored("ERROR: " + text, cls.color_error))
        if errorcode != 0:
            exit(errorcode)

    @classmethod
    def Info(cls, text, color=None):
        """
        Print the text in orange on the screen with the prefix INFO:"
        """
        if color is None:
            color = cls.color_info
        if text.startswith("\n"):
            print("")
            cls.Info(text[1:], color)
            return
        print(cls.colored("INFO: " + text, color))

    @classmethod
    def Warning(cls, text):
        """
        Print the text in orange on the screen with the prefix WARNING:"
        """
        print(cls.colored("WARNING: " + text, cls.color_info))

    @classmethod
    def InfoStr(cls, text, prefix=True):
        """
        Returns an info-str as printed by the Info method
        """
        if prefix:
            return cls.colored("INFO: " + text, cls.color_info)
        else:
            return cls.colored(text, cls.color_info)

    @classmethod
    def colored(cls, text, color):
        if color == "orange":
            text = "\033[93m" + text + "\033[0m"
        if color == "green":
            text = "\033[92m" + text + "\033[0m"
        if color == "red":
            text = "\033[91m" + text + "\033[0m"
        return text

    @classmethod
    def Indent(cls, text, nspace_first, nspace_following=None):
        """
        print a text indented and add line breaks at the terminal width
        """
        if nspace_following is None:
            nspace_following = nspace_first

        # split into single lines
        if text.startswith("\n"):
            print("")
            cls.Indent(text[1:], nspace_first, nspace_following)
            return
        lines = text.splitlines()
        if len(lines) > 1:
            cls.Indent(lines[0], nspace_first, nspace_following)
            for i in range(1, len(lines)):
                cls.Indent(lines[i], nspace_following)
            return

        # try to get the terminal size
        if cls.cols is None:
            pipe = subprocess.Popen(
                "stty size",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            output = str.join("", pipe.stdout.readlines())
            code = pipe.wait()
            if code == 0:
                cls.rows, cls.cols = map(lambda x: int(x), output.split())
            else:
                cls.cols = 0

        # how many space is available?
        avcols = cls.cols - nspace_first
        # terminal is to small?
        if avcols < 10 or len(text) <= avcols:
            print("%s%s" % (" " * nspace_first, text))
        else:
            print("%s%s" % (" " * nspace_first, text[0:avcols]))
            cls.Indent(text[avcols:], nspace_following)



#!/usr/bin/env python

from pathlib import Path
from setuptools.command.install import install
from setuptools import setup, find_packages
import shutil
import sys
from tempfile import TemporaryDirectory

from deploy import find_version, read, Installer, get_data_dirs


COMMANDS = ["databrowser", "esgf", "history", "plugin", "crawl-my-data"]
this_dir = Path(__file__).parent
COMPLETION_DIR = this_dir / "assets" / "completions"


def prep_tcsh_completion(tempdir):
    """Create completion scripts for tcsh."""

    script_dir = COMPLETION_DIR.relative_to(this_dir) / "tcsh"
    completion_helper = (
        Path(sys.prefix)
        / "share"
        / "tcsh-completion"
        / "helpers"
        / "tcsh-completion.bash"
    )
    script_paths = [
        (
            str(Path("share") / "tcsh-completion" / "helpers"),
            [str(script_dir / f"{completion_helper.name}")],
        ),
        (
            str(Path("share") / "tcsh-completion" / "completion"),
            [str(tempdir.relative_to(this_dir) / "freva")],
        ),
    ]
    completion_script = [
        "set autoexpand",
        f"complete freva 'p,*,`bash {completion_helper} \"${{COMMAND_LINE}}\"`,'",
    ]
    for cmd in COMMANDS:
        completion_script.append(
            f"complete {cmd} 'p,*,`bash {completion_helper} \"${{COMMAND_LINE}}\"`,'"
        )
    with (tempdir / "freva").open("w") as f:
        f.write("\n".join(completion_script))
    return script_paths


def gather_completion_scripts(tempdir):
    """Gather all data_files related to shell completion scripts."""

    # prefix = get_data_dirs(sys.prefix
    shells = {
        "zsh": Path("share") / "zsh" / "site-functions",
        "bash": Path("share") / "bash-completion" / "completions",
        "fish": Path("share") / "fish" / "completions",
    }
    data_files = []
    for shell, target_path in shells.items():
        comp_files = [
            str(f.relative_to(this_dir)) for f in (COMPLETION_DIR / shell).rglob("*")
        ]
        data_files.append((str(target_path), comp_files))
    return data_files + prep_tcsh_completion(tempdir)


class InstallCommand(install):
    """Customized setuptools install command."""

    def run(self):
        install.run(self)
        config_dir, data_dir, _ = get_data_dirs(self.prefix, self.user)
        with TemporaryDirectory(dir=COMPLETION_DIR) as td:
            data_files = gather_completion_scripts(Path(td))
            for target_dir, sources in data_files:
                full_target_dir = data_dir / target_dir
                for source in sources:
                    full_source = full_target_dir / Path(source).name
                    full_source.parent.mkdir(exist_ok=True, parents=True)
                    with full_source.open("w") as target_file:
                        with open(source, "r") as source_file:
                            target_file.write(source_file.read())
        if not (config_dir / "drs_config.toml").exists():
            shutil.copy(
                str(this_dir / "assets" / "drs_config.toml"),
                str(config_dir / "drs_config.toml"),
            )
        Installer.create_loadscript(self.prefix, bool(self.user))


def get_data_files():
    this_dir = Path(__file__).parent
    asset_dir = this_dir / "assets"
    dirs = [d for d in asset_dir.rglob("*") if d.is_dir()]
    files = []
    for d in dirs:
        target_dir = d.relative_to(this_dir)
        add_files = [str(f.relative_to(this_dir)) for f in d.rglob("*") if f.is_file()]
        if add_files:
            files.append((str(target_dir), add_files))
    files.append(("", ["deploy.py"]))
    return files


entry_points = ["freva = freva.cli:main"]
for cmd in COMMANDS:
    entry_points.append(f"freva-{cmd} = freva.cli.{cmd.replace('-', '_')}:main")
setup(
    name="freva",
    version=find_version("src/evaluation_system", "__init__.py"),
    author="German Climate Computing Centre (DKRZ)",
    maintainer="Climate Informatics and Technology (CLINT)",
    description="Free Evaluation and Analysis Framework (Freva) ",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    include_package_data=True,
    license="BSD-3-Clause",
    packages=find_packages("src"),
    package_dir={"": "src"},
    cmdclass={"install": InstallCommand},
    install_requires=[
        "appdirs",
        "GitPython",
        "dask",
        "Django",
        "humanize",
        "h5netcdf",
        "lazy-import",
        "mysqlclient",
        "netCDF4",
        "numpy",
        "pymysql",
        "pandas",
        "Pillow",
        "PyPDF2",
        "requests",
        "toml",
        "toolz",
        "typing_extensions",
        "xarray",
    ],
    setup_requires=["appdirs"],
    extras_require={
        "docs": [
            "cartopy",
            "cftime",
            "cf_xarray",
            "ipython",
            "nbsphinx",
            "pint",
            "pint-xarray",
            "recommonmark",
            "sphinx",
            "sphinxcontrib_github_alt",
            "sphinx-execute-code-python3",
            "sphinx-rtd-theme",
            "xarray",
            "h5netcdf",
        ],
        "test": [
            "allure-pytest",
            "black",
            "h5netcdf",
            "mock",
            "mypy",
            "nbval",
            "nbformat",
            "pep257",
            "pytest",
            "pytest-html",
            "pytest-env",
            "pytest-cov",
            "python-swiftclient",
            "requests_mock",
            "testpath",
            "types-mock",
            "types-requests",
            "types-toml",
            "types-urllib3",
            "types-toml",
            "types-requests",
        ],
    },
    entry_points={"console_scripts": entry_points},
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
    ],
)



"""
.. moduleauthor:: christopher kadow / Sebastian Illing
.. first version written by estanislao gonzalez

The module encapsulates all methods for accessing files on the system.
These are mainly model and observational and reanalysis data.
"""
from __future__ import annotations
from typing import (
    Optional,
    Generator,
    List,
    Union,
    Any,
    ClassVar,
    cast,
)
from typing_extensions import Literal, TypedDict
from dataclasses import dataclass, field

import json
from pathlib import Path
import os
import logging
from evaluation_system.misc import config

log = logging.getLogger(__name__)

Activity = str
"""Represents a type of data collection activity for DRS (see Activity from the DRS spec).

    This doesn't prevent typing issues beyond what you'd get with `str` but  I think
    it makes the functions that deal with Activities more clear than simply using `str`
"""

ACTIVITY_BASELINE0: Activity = "baseline0"


@dataclass
class DRSStructure:
    root_dir: str
    """Directory from where this files are to be found. Put through `expanduser` to
        expand `~` then `absolute`.
    """
    parts_dir: list[str]
    """List of subdirectory category names the values they refer to
        (e.g. ['model', 'experiment']).
    """
    parts_file_name: list[str]
    """Elements composing the file name (no ".nc" though)."""
    parts_time: str
    """Describes how the time part of the filename is formed."""
    dataset: Activity
    """same value as this key structure (for reverse traverse)"""
    defaults: dict[str, str] = field(default_factory=dict)
    """list with values that "shouldn't" be required to be changed (e.g. for
        observations, project=obs4MIPS)
    """

    def __post_init__(self) -> None:
        self.root_dir = str(Path(self.root_dir).expanduser().absolute())
        self.parts_dataset: list[str] = [
            "project",
            "product",
            "institute",
            "model",
            "experiment",
            "time_frequency",
            "realm",
            "cmor_table",
            "ensemble",
            "variable",
        ]
        self.parts_versioned_dataset: list[str] = [
            "project",
            "product",
            "institute",
            "model",
            "experiment",
            "time_frequency",
            "realm",
            "cmor_table",
            "ensemble",
            "version",
            "variable",
        ]

    @classmethod
    def from_dict(cls, dataset: str, drs_dict: dict[str, Any]) -> DRSStructure:
        """Creates a DRSStructure from the given dict

        Parameters
        ----------
        dataset:
            Name of the DRSStructure (key in the drs_config.toml file)
        drs_dict:
            Dictionary containing all DRSStructure fields

        Returns
        -------
        DRSStructure
            DRSStructure created from dict
        """
        d = cls(
            root_dir=drs_dict["root_dir"],
            parts_dir=drs_dict["parts_dir"],
            parts_file_name=drs_dict["parts_file_name"],
            parts_time=drs_dict.get("parts_time", ""),
            dataset=dataset,
        )
        if "defaults" in drs_dict:
            d.defaults = drs_dict["defaults"]
        return d


class FileComponents(TypedDict):
    root_dir: str
    # This is incorrect, parts is actually `dict[str, Optional[str]]` due to a check in
    # from_path where it will insert a None under certain conditions.
    # Fixing this would require changing a lot of code that currently seems to be
    # working since Python's types are nullable by default so I guess it's fine until we
    # want to start cleaning up ignores.
    # The code also makes assumptions about the contents of parts like assuming the
    # presence of a file_name key. I'm not sure this is worth capturing right now since
    # I think it relates heavily to the DRS structure and thus the DRSStructure type
    # which similarly doesn't have much encoded into its type.
    parts: dict[str, str]


class DRSFile:
    """Represents a file that follows the
    `DRS standard <https://pcmdi.llnl.gov/mips/cmip5/docs/cmip5_data_reference_syntax.pdf>`_.
    """

    # Lazy initialized in find_structure_from_path
    DRS_STRUCTURE_PATH_TYPE: ClassVar[Optional[dict[str, Activity]]] = None
    DRS_STRUCTURE: ClassVar[Optional[dict[Activity, DRSStructure]]] = None

    def __init__(
        self,
        file_dict: Optional[FileComponents] = None,
        drs_structure: Activity = ACTIVITY_BASELINE0,
    ):
        """Creates a DRSfile out of the dictionary containing information
            about the file or from scratch.

        Parameters
        ----------
        file_dict
            dictionary with the DRS component values and keys from which
            this file will be initialized.
        drs_structure
            Which structure is going to be used with this file. Expected
            to be a key value from `DRSFile.DRS_STRUCTURE`
        """
        self.drs_structure = drs_structure
        if not file_dict:
            file_dict = {
                "root_dir": "",
                "parts": {},
            }
        self.dict = file_dict
        # trim the last slash if present in root_dir
        if self.dict["root_dir"]:
            self.dict["root_dir"] = str(
                Path(self.dict["root_dir"]).expanduser().absolute()
            )

    def __repr__(self) -> str:  # pragma: no cover
        """Get the JSON representation.

        This can be used for creating a copy via `from_json`.

        Returns
        -------
        str
            json representation of this object
        """
        return self.to_json()

    def __str__(self) -> str:  # pragma: no cover
        """Return the path of the file.

        See `to_path` for more info.

        Returns
        -------
        str
            path of the file
        """
        return self.to_path()

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, DRSFile):
            return NotImplemented
        return self.to_path() < other.to_path()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DRSFile):
            return False
        return self.to_path() == other.to_path()

    def to_json(self) -> str:
        """:returns: (str) the json representation of the dictionary encapsulating the DRS components of this file."""
        return json.dumps(self.dict)

    def to_path(self) -> str:
        """Return the path of the file.

        This returns the full path to the file as described by the DRS
        components. The file is not required to exist.

        Returns
        -------
        str
            The path to the file

        Raises
        ------
        KeyError
            If it can't construct the path because information is missing
            in the DRS components.
        """
        # TODO: check if construction is complete and therefore can succeed
        result = self.dict["root_dir"]
        for key in self.get_drs_structure().parts_dir:
            if key not in self.dict["parts"]:
                raise KeyError("Can't construct path as key %s is missing." % key)
            result = os.path.join(result, self.dict["parts"][key])
        return os.path.join(result, self.dict["parts"]["file_name"])

    def to_dataset(self, versioned: bool = False, to_path: bool = False) -> str:
        """Returns dataset information.

        This will return either the dataset's identifier or the path to
        the dataset depending on the value of `to_path`.

        Parameters
        ----------
        versioned
            If the dataset should contain information about the version.
            Note that not all DRS structures are versioned, so in those
            cases where there is just no version information this makes
            no difference.
        to_path
            If true return the path to the dataset, otherwise returns the
            dataset identifier.

        Returns
        -------
        str
            Either the path to the dataset or the dataset identifier

        Raises
        ------
        ValueError
            If `versioned` is True but the structure is not versioned.
        """
        result = []
        structure = self.get_drs_structure()
        if versioned and self.versioned:
            iter_parts = structure.parts_versioned_dataset
        elif versioned:
            raise ValueError(f"{self.drs_structure} is not versioned!")
        else:
            iter_parts = structure.parts_dataset

        for key in iter_parts:
            if key in self.dict["parts"]:
                result.append(self.dict["parts"][key])
            elif key in structure.defaults:
                result.append(structure.defaults[key])
        if to_path:
            return os.path.join(structure.root_dir, os.sep.join(result))
        else:
            return ".".join(result)

    def to_dataset_path(self, versioned: bool = False) -> str:
        """Returns the path to the current dataset.

        We are assuming the dataset is a sub-path of all files in it.

        See `to_dataset` for more information.

        Parameters
        ----------
        versioned
            If the dataset should contain information about the version.

        Returns
        -------
        str
            The path to the dataset
        """
        return self.to_dataset(versioned=versioned, to_path=True)

    @property
    def versioned(self) -> bool:
        """If this file is from a DRS structure that is versioned.

        Returns
        -------
        bool
            True is the dataset is versioned, False otherwise
        """
        return self.dict["parts"].get("version") is not None

    @property
    def version(self) -> Optional[str]:
        """Returns the dataset version of this file

        This returns the version which this file is a part of or None if
        the dataset is not versioned. Note that this is the version of the
        dataset and not of the file since the DRS does not have file
        versions.

        Returns
        -------
        Optional[str]
            The version of the dataset or None if not versioned
        """
        return self.dict["parts"].get("version")

    @staticmethod
    def _get_structure_prefix_map() -> dict[str, Activity]:
        """Returns reversed map of root_dir to Activity name.

        This will lazily initialize the map if it doesn't already exist.

        Returns
        -------
        dict[str, Activity]
            Map of root_dir to Activity name
        """

        if DRSFile.DRS_STRUCTURE_PATH_TYPE is None:
            DRSFile._load_structure_definitions()
        # ignored due to lazy initialization issue
        return DRSFile.DRS_STRUCTURE_PATH_TYPE  # type: ignore [return-value]

    @staticmethod
    def find_structure_from_path(
        file_path: str, allow_multiples: bool = False
    ) -> list[Activity]:
        """Return all DRS structures that might be applicable.

        This is resolved by checking if the prefix of any structure paths
        matches that of the given file path. Parsing is not done, so it
        might still fail. This just guarantees that only the structures
        returned here *might* work.

        Parameters
        ----------
        file_path
            Full path to a file, whose drs structure is being searched for.
        allow_multiples
            If true returns a list with all possible structures, otherwise
            returns the first match found.
        Returns
        -------
        Union[Activity, List[Activity]]
            The name of the drs struct(s) that can be used to parse this path.

        Raises
        ------
        ValueError
            If `file_path` does not correspond with any DRS structure.
        """
        structures = []
        for path_prefix, st_type in DRSFile._get_structure_prefix_map().items():
            if file_path.startswith(path_prefix):
                if allow_multiples:
                    structures.append(st_type)
                else:
                    return [st_type]
        if not structures:
            raise ValueError(f"Unrecognized DRS structure in path {file_path}")
        else:
            return structures

    @staticmethod
    def find_structure_in_path(
        dir_path: str, allow_multiples: bool = False
    ) -> Union[Activity, list[Activity]]:
        """Return all DRS structures that might be applicable.

        See `find_structure_in_path` for more information

        Parameters
        ----------
        file_path
            Path to a directory which might contain DRS files.
        allow_multiples
            If true returns a list with all possible structures, otherwise
            returns the first match found.

        Returns
        -------
        Union[Activity, List[Activity]]
            The name of the drs struct(s) that can be used to parse this path.

        Raises
        ------
        ValueError
            If `dir_path` does not correspond with any DRS structure.
        """
        structures = []
        for path_prefix, st_type in DRSFile._get_structure_prefix_map().items():
            if dir_path.startswith(path_prefix):
                if allow_multiples:
                    structures.append(st_type)
                else:
                    return st_type
        if not structures:
            raise ValueError(f"No DRS structure found in {dir_path}.")
        else:
            return structures

    @staticmethod
    def from_path(path: os.PathLike, activity: Optional[Activity] = None) -> DRSFile:
        """Extract a DRSFile object out of a path.

        Parameters
        ----------
        path
            Path to a file that is part of the ``drs_structure``.
        drs_structure
            Which structure is going to be used with this file.

        Returns
        -------
        DRSFile
            File extracted from path

        Raises
        ------
        ValueError
            If the given path cannot be used in the given DRS Structure
            or any configured structure if `activity` is None.
        """
        path = Path(path).expanduser().absolute()
        if activity is None:
            activity = DRSFile.find_structure_from_path(str(path))[0]
        structure = DRSFile._get_drs_structure(activity)

        try:
            parts = path.parent.relative_to(structure.root_dir).parts
        except ValueError as error:
            raise ValueError(
                f"File {path} does not correspond to {activity}"
            ) from error

        # check the number of parts
        if len(parts) != len(structure.parts_dir):
            raise ValueError(
                (
                    f"Can't parse this path. Expected {len(structure.parts_dir)} "
                    f"elements but got {len(parts)}."
                )
            )

        # first the dir
        result: FileComponents = {
            "root_dir": structure.root_dir,
            "parts": dict(zip(structure.parts_dir, parts)),
        }
        result["parts"]["file_name"] = path.name
        # split file name
        # (extract .nc before splitting)
        file_name_parts: list[str] = path.with_suffix("").name.split("_")
        if (
            len(file_name_parts) == len(structure.parts_file_name) - 1
            and "fx" in file_name_parts
        ):
            # no time
            file_name_parts.append("")
        if len(file_name_parts) != len(structure.parts_file_name):
            raise ValueError(
                f"File {path} does not follow the expected naming scheme for {activity}"
            )
        for key, value in dict(zip(structure.parts_file_name, file_name_parts)).items():
            result["parts"].setdefault(key, value)
        return DRSFile(result, drs_structure=activity)

    def get_drs_structure(self) -> DRSStructure:
        """Returns the DRS structure used by this file.

        Returns
        -------
        DRSStructure
            The `DRS_STRUCTURE` used by this file."""
        return DRSFile._get_drs_structure(self.drs_structure)

    @staticmethod
    def _get_drs_structure(
        drs_structure: Activity = ACTIVITY_BASELINE0,
    ) -> DRSStructure:
        """Gets the DRSStructure associated with the given Activity

        Parameters
        ----------
        drs_structure
            Name of a DRS structure

        Returns
        -------
        DRSStructure
            The DRS structure of the requested type.

        Raises
        ------
        ValueError
            If there's no such DRS structure.
        """
        if DRSFile.DRS_STRUCTURE is None:
            DRSFile._load_structure_definitions()
        # these ignores are because mypy doesn't go into `_load_structure_definitions`
        # to see that DRS_STRUCTURE is initialized before being used and there isn't a
        # way to force that that wouldn't make the code much worse
        if drs_structure not in DRSFile.DRS_STRUCTURE:  # type: ignore [operator]
            raise ValueError(f"Unknown DRS structure {drs_structure}")
        return DRSFile.DRS_STRUCTURE[drs_structure]  # type: ignore [index]

    @staticmethod
    def from_dict(
        file_dict: FileComponents, drs_structure: Activity = ACTIVITY_BASELINE0
    ) -> DRSFile:
        """Creates a DRSFile based off given dict

        Parameters
        ----------
        file_dict
            Dictionary with the DRS components.
        drs_structure
            Name of a DRS structure

        Returns
        -------
        DRSFile
            The DRSFile generated from the given dictionary and DRS
            structure name
        """
        return DRSFile(file_dict, drs_structure=drs_structure)

    @staticmethod
    def from_json(
        json_str: str, drs_structure: Activity = ACTIVITY_BASELINE0
    ) -> DRSFile:
        """Creates a DRSFile based off the given JSON string.

        Parameters
        ----------
        json_str
            JSON representation of the DRS components. Like the result
            from calling :class:`DRSFile.to_json`.
        drs_structure
            Name of a DRS structure

        Returns
        -------
        DRSFile
            The DRSFile generated from the given dictionary and DRS
            structure name
        """
        return DRSFile.from_dict(json.loads(json_str), drs_structure=drs_structure)

    @staticmethod
    def solr_search(
        drs_structure: Optional[str] = None,
        latest_version: bool = True,
        path_only: bool = False,
        batch_size: int = 10000,
        **partial_dict: Any,
    ) -> Generator[Union[Any, DRSFile], None, None]:
        """Search for files by relying on a Solr Index.*'.

        Parameters
        ----------
        drs_structure
            Name of a DRS structure (key of :class:`DRSFile.DRS_STRUCTURE`). This isn't mandatory anymore.
        latest_version
            If this should be only the latest version available.
        path_only
            If true returns a string with the path only, otherwise a complete DRSFile object.
        batch_size
            The size of the number of results that will be returned by each Solr call.
        partial_dict
            A dictionary with some DRS components representing the query.

        Returns
        -------
        Generator[Union[Any, DRSFile], None, None]
            Generator returning matching files
        """
        from evaluation_system.model.solr import SolrFindFiles

        if drs_structure is not None:
            partial_dict["dataset"] = drs_structure
        if path_only:
            for path in SolrFindFiles.search(batch_size=batch_size, **partial_dict):
                yield path
        else:
            for path in SolrFindFiles.search(batch_size=batch_size, **partial_dict):
                yield DRSFile.from_path(path)

    @staticmethod
    def _load_structure_definitions() -> None:
        """Loads DRSStructure definitions from the config.

        This handles the initialization of `DRS_STRUCTURE_PATH_TYPE` and
        `DRS_STRUCTURE`.
        """
        DRSFile.DRS_STRUCTURE_PATH_TYPE = {}
        DRSFile.DRS_STRUCTURE = {}

        conf = config.get_drs_config()

        for a, structure in conf.items():
            activity = Activity(a)
            try:
                ds = DRSStructure.from_dict(activity, structure)
            except KeyError:
                # Missing information, skip this entry
                continue
            DRSFile.DRS_STRUCTURE[activity] = ds

            path_prefix = ds.root_dir
            for part in ds.parts_dir:
                # if we have more info use it to generate a unique root path
                # (e.g. cmip5 and baseline0 or baseline1 share the same root path!)
                if part in ds.defaults and all(
                    [char not in ds.defaults[part] for char in "*?"]
                ):
                    # but only use it if it's a plain string, no globing.
                    path_prefix += os.sep + ds.defaults[part]
                else:
                    break
            DRSFile.DRS_STRUCTURE_PATH_TYPE[path_prefix] = activity



"""
Created on 17.5.2016

@author: Sebastian Illing
"""
import subprocess
import os
from pathlib import Path
import tempfile
import shutil

import pytest

DUMMY_USER = {"pw_name": "someone"}


def test_dummy_user(temp_user, dummy_settings, dummy_env):
    """Be sure the dummy user is created as expected"""
    dummy_name = "non-existing name"
    from evaluation_system.model.user import User
    from evaluation_system.tests.mocks.dummy import DummyUser

    with pytest.raises(Exception):
        DummyUser(random_home=True, pw_dir="anything")

    with DummyUser(random_home=True, pw_name=dummy_name) as d_user:
        # populate the test directory as required
        d_user.prepareDir()
        assert dummy_name == d_user.getName()
        cfg_file = os.path.join(d_user.getUserBaseDir(), User.EVAL_SYS_CONFIG)

        # check configuration file writing
        assert not os.path.isfile(cfg_file)
        cnfg = d_user.getUserConfig()
        cnfg.add_section("test_section")
        cnfg.set("test_section", "some_key", "a text value\nwith many\nlines!")
        d_user.writeConfig()
        assert os.path.isfile(cfg_file)
        # check configuration file reading
        with open(cfg_file, "w") as fp:
            ##Note multi line values in the configuration file need to be indented...
            fp.write(
                "[test2]\nkey1 = 42\nkey2=Some\n\tmany\n\tlines=2\nkey3=%(key1)s0\n"
            )
        cnfg = d_user.reloadConfig()
        assert cnfg.getint("test2", "key1") == 42
        # ...but the indentation disappears when returned directly
        assert cnfg.get("test2", "key2") == "Some\nmany\nlines=2"
        assert cnfg.getint("test2", "key3") == 420


def test_getters(dummy_settings, dummy_env):
    """Test the object creation and some basic return functions"""
    from evaluation_system.model.user import User
    from evaluation_system.misc import config
    from evaluation_system.tests.mocks.dummy import DummyUser

    try:
        config._config[config.DIRECTORY_STRUCTURE_TYPE] = "local"
        with DummyUser(random_home=True, pw_name="someone") as temp_user:
            assert DUMMY_USER["pw_name"] == temp_user.getName()
            assert temp_user.getUserHome().startswith(tempfile.gettempdir())
            assert os.getuid() == temp_user.getUserID()
            print(config.get("directory_structure_type"))
            db = temp_user.getUserDB()
            assert db is not None
            baseDir = "/".join([temp_user.getUserHome(), config.get(config.BASE_DIR)])
            assert baseDir == temp_user.getUserBaseDir()
            tool1_cnfDir = os.path.join(baseDir, User.CONFIG_DIR, "tool1")
            tool1_chDir = os.path.join(baseDir, User.CACHE_DIR, "tool1")
            tool1_outDir = os.path.join(baseDir, User.OUTPUT_DIR, "tool1")
            tool1_plotDir = os.path.join(baseDir, User.PLOTS_DIR, "tool1")
            assert temp_user.getUserScratch() == "/tmp/scratch/%s" % temp_user.getName()
            # check we get the configuration directory of the given tool
            assert tool1_cnfDir == temp_user.getUserConfigDir("tool1")
            assert tool1_chDir == temp_user.getUserCacheDir("tool1")
            assert tool1_outDir == temp_user.getUserOutputDir("tool1")
            assert tool1_plotDir == temp_user.getUserPlotsDir("tool1")
            # check we get the general directory of the tools (should be the parent of the previous one)
            assert os.path.dirname(tool1_cnfDir) == temp_user.getUserConfigDir()
            assert os.path.dirname(tool1_chDir) == temp_user.getUserCacheDir()
            assert os.path.dirname(tool1_outDir) == temp_user.getUserOutputDir()
            assert os.path.dirname(tool1_plotDir) == temp_user.getUserPlotsDir()
    finally:
        config.reloadConfiguration()


def test_directory_creation(temp_user, dummy_env):
    """
    This tests assures we always know what is being created
    in the framework directory
    """
    from evaluation_system.tests.mocks.dummy import DummyUser

    # assure we have a temp directory as HOME for testing
    testUserDir = tempfile.mkdtemp("_es_userdir")
    with DummyUser(pw_dir=testUserDir) as testUser:
        # assure we have a home directory setup
        assert os.path.isdir(testUser.getUserHome())
        baseDir = testUser.getUserBaseDir()
        # check home is completely created
        # self.assertFalse(os.path.isdir(baseDir)) Basic dir structure is created when running a tool because of
        # history, so this doesn't apply anymore
        testUser.prepareDir()
        assert os.path.isdir(baseDir)
        created_dirs = [
            testUser.getUserConfigDir(),
            testUser.getUserCacheDir(),
            testUser.getUserOutputDir(),
            testUser.getUserPlotsDir(),
        ]
        for directory in created_dirs:
            assert os.path.isdir(directory)


def test_directory_creation2(temp_user, dummy_env):
    try:

        testUser = temp_user
        # assure we have a home directory setup
        assert os.path.isdir(testUser.getUserHome())
        dir1 = temp_user.getUserConfigDir("test_tool", create=False)
        assert not os.path.isdir(dir1)
        dir2 = temp_user.getUserConfigDir("test_tool", create=True)
        assert dir1 == dir2
        assert os.path.isdir(dir1)
    finally:
        shutil.rmtree(Path(temp_user.getUserBaseDir()).parent)


def test_central_directory_Creation(temp_dir, dummy_settings, dummy_env):
    from evaluation_system.tests.mocks.dummy import DummyUser
    from evaluation_system.misc import config

    try:
        config._config[config.BASE_DIR_LOCATION] = str(temp_dir)
        config._config[
            config.DIRECTORY_STRUCTURE_TYPE
        ] = config.DIRECTORY_STRUCTURE.CENTRAL
        with DummyUser(random_home=False, **DUMMY_USER) as temp_user:
            dir1 = temp_user.getUserBaseDir()
            assert dir1 == os.path.join(
                config.get(config.BASE_DIR_LOCATION),
                config.get(config.BASE_DIR),
                str(temp_user.getName()),
            )
            dir2 = temp_user.getUserOutputDir("sometool")
            assert dir2 == os.path.join(
                config.get(config.BASE_DIR_LOCATION),
                config.get(config.BASE_DIR),
                str(temp_user.getName()),
                User.OUTPUT_DIR,
                "sometool",
            )
    finally:
        config.reloadConfiguration()
        # if os.path.isdir(temp_dir) and temp_dir.startswith(tempfile.gettempdir()):
        #    #make sure the home is a temporary one!!!
        #    shutil.rmtree(temp_dir)


def test_config_file(temp_user, dummy_settings):
    tool = "test_tool"
    assert temp_user.getUserConfigDir(
        tool
    ) + "/%s.conf" % tool == temp_user.getUserToolConfig(tool)



"""Tests for ingesting user data."""
from __future__ import annotations
from pathlib import Path
from tempfile import TemporaryDirectory, NamedTemporaryFile
from typing import Generator, Tuple

import cftime
import pytest
import numpy as np
import xarray as xr


def create_data(
    num_files: int,
    variable_names: list[str],
    chunk_size: Tuple[int, ...],
    dims: Tuple[str, ...],
) -> Generator[Path, None, None]:
    """Create a netcdf dataset."""
    coords = {d: np.ones(chunk_size[n]) for (n, d) in enumerate(dims)}
    data = xr.Dataset()
    attrs = {}
    ntstep = 0
    with TemporaryDirectory() as temp_dir:
        for nfile in range(num_files):
            if "time" in dims:
                units = "seconds since 1970-01-01 00:00:00"
                coords["time"] = cftime.num2date(
                    np.array([ntstep + (j * 3600) for j in range(chunk_size[0])]),
                    units=units,
                )
                data["time"] = xr.DataArray(
                    data=coords["time"],
                    dims=("time",),
                    coords={"time": coords["time"]},
                    name="time",
                )
                attrs["frequency"] = "not valid"
            for variable_name in variable_names:
                dset = xr.DataArray(
                    np.zeros(chunk_size),
                    dims=dims,
                    coords=coords,
                    name=variable_name,
                )
                data[variable_name] = dset
                out_f = Path(temp_dir) / f"outfile_{nfile}.nc"
                data.set_coords(dims).to_netcdf(str(out_f))
            ntstep += chunk_size[0] * 3600
        yield Path(temp_dir)


@pytest.fixture(scope="function")
def valid_data_files() -> Generator[Path, None, None]:
    yield from create_data(
        10,
        ["tas", "time_bnds", "X", "rotated_pole", "longitude"],
        (10, 10, 10),
        ("time", "lat", "lon"),
    )


@pytest.fixture(scope="function")
def invalid_data_files() -> Generator[Path, None, None]:
    yield from create_data(1, ["tas", "foo"], (10, 10), ("lat", "lon"))


def test_invalid_data_files(invalid_data_files: Path) -> Path:
    from evaluation_system.api.user_data import DataReader

    in_file = list(invalid_data_files.rglob("*.*"))[0]
    data_reader = DataReader(invalid_data_files)
    with pytest.raises(ValueError):
        data_reader.get_metadata(in_file)
    not_a_nc_file = invalid_data_files / "not_a_nc_file.nc"
    not_a_nc_file.touch()
    data_reader = DataReader(not_a_nc_file)
    with pytest.raises(ValueError):
        data_reader.get_metadata(not_a_nc_file)


def test_add_valid_data(valid_data_files: Path) -> None:
    from evaluation_system.api.user_data import DataReader

    in_file = list(valid_data_files.rglob("*.*"))[0]
    data_reader = DataReader(valid_data_files)
    data = data_reader.get_metadata(in_file)
    assert data["variable"] == "tas"
    assert data["time_frequency"] == "hr"
    data_reader = DataReader(valid_data_files, variable="foo")
    data = data_reader.get_metadata(in_file)
    assert data["variable"] == "foo"
    assert data["time_frequency"] == "hr"
    with xr.open_dataset(str(in_file)) as dset:
        dset = dset.isel(time=0).load()
    dset.to_netcdf(in_file)
    data = data_reader.get_metadata(in_file)
    assert data["variable"] == "foo"
    assert data["time_frequency"] == "fx"


def test_get_time_frequency(valid_data_files: Path) -> None:
    from evaluation_system.api.user_data import DataReader

    data_reader = DataReader("foo/bar.nc")
    hour = 3600
    day = hour * 24
    mon = day * 30
    year = day * 360
    assert data_reader.get_time_frequency(0, "foo") == "fx"
    assert data_reader.get_time_frequency(0, "hr") == "hr"
    assert data_reader.get_time_frequency(60) == "subhr"
    assert data_reader.get_time_frequency(hour) == "hr"
    assert data_reader.get_time_frequency(day) == "day"
    assert data_reader.get_time_frequency(15 * day) == "sem"
    assert data_reader.get_time_frequency(mon) == "mon"
    assert data_reader.get_time_frequency(year) == "yr"


def test_get_file_name_from_metadata(valid_data_files: Path) -> None:
    from evaluation_system.api.user_data import DataReader

    in_file = list(valid_data_files.rglob("*.*"))[0]
    defaults = dict(
        project="foo",
        product="fumanshu",
        institute="tong",
        model="mrfu",
        experiment="foo-boo",
        cmor_table="foo",
        ensemble="bar",
        realm="foo-kingdom",
        version="v0",
    )
    data_reader = DataReader(valid_data_files, **defaults)
    file_part = "foo/fumanshu/tong/mrfu/foo-boo/hr/foo-kingdom/foo/bar/tas"
    assert file_part in str(data_reader.file_name_from_metdata(in_file).parent)
    defaults["variable"] = "foob"
    defaults["time_frequency"] = "3h"
    data_reader = DataReader(valid_data_files, **defaults)
    file_part = "foo/fumanshu/tong/mrfu/foo-boo/3h/foo-kingdom/foo/bar/foob"
    assert file_part in str(data_reader.file_name_from_metdata(in_file).parent)
    defaults.pop("cmor_table")
    data_reader = DataReader(valid_data_files, **defaults)
    with pytest.raises(ValueError):
        data_reader.file_name_from_metdata(in_file)


def test_iter_data_files(valid_data_files: Path) -> None:
    from evaluation_system.api.user_data import DataReader

    input_files = list(valid_data_files.rglob("*.*"))
    not_a_nc_file = valid_data_files / "not_a_nc_file.txt"
    not_a_nc_file.touch()
    data_reader = DataReader(valid_data_files)
    files = [f for f in data_reader]
    data_reader = DataReader(valid_data_files / "*.nc")
    files = [f for f in data_reader]
    assert len(files) == len(input_files)
    data_reader = DataReader(valid_data_files / "*.txt")
    files = [f for f in data_reader]
    assert files == []
    data_reader = DataReader(not_a_nc_file)
    files = [f for f in data_reader]
    assert files == []


def test_link_my_data(dummy_crawl, dummy_plugin, valid_data_files):

    from evaluation_system.model.solr import SolrFindFiles

    input_files = list(valid_data_files.rglob("*.nc"))
    dummy_plugin.add_output_to_databrowser(valid_data_files)
    assert len(list(SolrFindFiles.search(latest_version=False))) == len(input_files)


def test_crawl_my_data(dummy_crawl, capsys, dummy_env, valid_data_files):
    from freva import crawl_my_data
    from evaluation_system.tests import run_cli
    from freva.cli.crawl_my_data import main as run
    from evaluation_system.misc.exceptions import ValidationError
    from evaluation_system.model.solr import SolrFindFiles

    run(["--data-type=fs"])
    captured = capsys.readouterr()
    assert "Status: crawling ..." in captured.out
    assert "ok" in captured.out
    assert len(list(SolrFindFiles.search())) == len(dummy_crawl)
    assert len(list(SolrFindFiles.search(latest_version=False))) == len(dummy_crawl)
    with pytest.raises(NotImplementedError):
        crawl_my_data(dtype="something")
    with pytest.raises(SystemExit):
        with pytest.raises(ValidationError):
            run_cli(["crawl-my-data", "/tmp/forbidden/folder"])


def test_wrong_datatype(dummy_crawl, capsys, dummy_env):

    from evaluation_system.tests import run_cli
    from evaluation_system.model.solr import SolrFindFiles

    dummy_crawl.append(dummy_crawl[0].parent / "more_info" / dummy_crawl[0].name)
    dummy_crawl[-1].parent.mkdir(exist_ok=True, parents=True)
    dummy_crawl[-1].touch()
    with pytest.raises(ValueError):
        run_cli(["crawl-my-data", "-d"])
    with pytest.raises(SystemExit):
        run_cli(["crawl-my-data"])
    captured = capsys.readouterr()
    assert "ValueError" in captured.err
    assert len(list(SolrFindFiles.search())) == 0
    assert len(list(SolrFindFiles.search(latest_version=False))) == 0


def test_validate_path(root_path_with_empty_config):
    from freva._crawl_my_data import _validate_user_dirs

    root_path_str = str(root_path_with_empty_config)
    print(root_path_str)
    assert _validate_user_dirs(root_path_str) == (root_path_with_empty_config,)



"""General Freva commandline argument parser."""

import argparse
import sys

from evaluation_system import __version__
from .utils import BaseParser
from typing import Optional, List

COMMAND = "freva"


class ArgParser(BaseParser):
    """Cmd argument parser class for main entry-point."""

    def __init__(self, argv: List[str]):

        epilog = f"""To get help for the individual sub-commands use:
        {COMMAND} <sub-command> --help
"""
        argv = argv or sys.argv[1:]
        sub_parsers = self.get_subcommand_parsers()
        try:
            if argv[0].strip("-") in sub_parsers:
                argv[0] = argv[0].strip("-")
        except IndexError:
            argv.append("-h")
        parser = argparse.ArgumentParser(
            prog=COMMAND,
            epilog=epilog,
            description="Free EVAluation system framework (freva)",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        parser.add_argument(
            "-V",
            "--version",
            action="version",
            version="%(prog)s {version}".format(version=__version__),
        )
        super().__init__(sub_parsers, parser)
        args = self.parse_args(argv)
        try:
            args.apply_func(args, **self.kwargs)
        except KeyboardInterrupt:
            print("KeyboardInterrupt, exiting", file=sys.stderr, flush=True)
            sys.exit(130)


def main(argv: Optional[List[str]] = None) -> None:
    """Wrapper for entrypoint script."""
    ArgParser(argv or sys.argv[1:])


if __name__ == "__main__":  # pragma: no cover
    main()



from evaluation_system.api.workload_manager.sge import SGEJob


def test_job_script(tmpdir):
    log_directory = tmpdir.strpath
    with SGEJob(
        cores=6,
        processes=2,
        memory="12GB",
        queue="my-queue",
        project="my-project",
        walltime="02:00:00",
        env_extra=["export MY_VAR=my_var"],
        job_extra=["-w e", "-m e"],
        log_directory=log_directory,
        resource_spec="h_vmem=12G,mem_req=12G",
    ) as cluster:
        job_script = cluster.job_script()

        for each in [
            "-q my-queue",
            "-P my-project",
            "-l h_rt=02:00:00",
            "export MY_VAR=my_var",
            "#$ -w e",
            "#$ -m e",
            "#$ -e {}".format(log_directory),
            "#$ -o {}".format(log_directory),
            "-l h_vmem=12G,mem_req=12G",
            "#$ -cwd",
            "#$ -j y",
        ]:
            assert each in job_script



import pytest


def test_get_cancel_command():
    from evaluation_system.api.workload_manager import cancel_command

    with pytest.raises(NotImplementedError):
        cancel_command("bla", 1000)
    assert "scancel" in cancel_command("slurm", 1000)



from __future__ import annotations
import argparse
from pathlib import Path
import sys
from typing import Optional, Any, cast, Union

from evaluation_system import __version__
import freva
from .utils import BaseCompleter, BaseParser

CLI = "EsgfCli"


class EsgfCli(BaseParser):
    """Class that constructs the ESGF Query Argument Parser."""

    desc = "Search/Download ESGF the data catalogue."

    def __init__(
        self,
        command: str = "freva",
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the esgf sub arg. parser."""
        subparser = parser or argparse.ArgumentParser(
            prog=f"{command}-esgf",
            description=self.desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        subparser.add_argument(
            "--datasets",
            default=False,
            action="store_true",
            help="List the name of the datasets instead of showing the urls.",
        )
        subparser.add_argument(
            "--show-facet",
            default=None,
            action="append",
            help=(
                "List all values for the given facet (might be "
                "defined multiple times). The results show the possible "
                "values of the selected facet according to the given "
                "constraints and the number of *datasets* (not files) "
                "that selecting such value as a constraint will result "
                "(faceted search)"
            ),
        )
        subparser.add_argument(
            "--opendap",
            default=False,
            action="store_true",
            help="Show opendap endpoints instead of http onse.",
        )
        subparser.add_argument(
            "--gridftp",
            default=False,
            action="store_true",
            help=(
                "Show gridftp endpoints instead of the http default "
                "ones (or skip them if none found)"
            ),
        )
        subparser.add_argument(
            "--download-script",
            default=None,
            type=Path,
            help=(
                "Download wget_script for getting the files "
                "instead of displaying anything (only http) "
            ),
        )
        subparser.add_argument(
            "--query",
            default=None,
            type=str,
            help=("Display results from <list> queried fields"),
        )
        subparser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        subparser.add_argument(
            "facets", nargs="*", help="Search facet(s)", type=str, metavar="facets"
        )
        self.parser = subparser
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(
        args: argparse.Namespace, other_args: Optional[list[str]] = None, **kwargs: Any
    ) -> None:
        """Call the esgf command and print the results."""
        facets: dict[str, Union[list[str], str]] = {}
        _facets = BaseCompleter.arg_to_dict(args.facets)
        _ = kwargs.pop("facets", None)
        for key, val in _facets.items():
            if len(val) == 1:
                facets[key] = val[0]
            else:
                facets[key] = val
        merged_args = cast(Any, {**kwargs, **facets})
        out = freva.esgf(**merged_args)
        if not out:
            return
        if args.datasets:
            print("\n".join([f"{d[0]} - version: {d[1]}" for d in out]))
        elif args.query:
            if len(args.query.split(",")) > 1:
                print("\n".join([str(out)]))
            else:
                print("\n".join([str(d) for d in list(out)]))
        elif args.show_facet:
            for facet_key in sorted(out):
                if len(out[facet_key]) == 0:
                    values = "<No Results>"
                else:
                    values = "\n\t".join(
                        [
                            "%s: %s" % (k, out[facet_key][k])
                            for k in sorted(out[facet_key])
                        ]
                    )
                    print("[%s]\n\t%s" % (facet_key, values))
        elif args.download_script:
            print(out)
        else:
            print("\n".join([str(d) for d in out]))


def main(argv: Optional[list[str]] = None) -> None:
    """Wrapper for entry point script."""
    cli = EsgfCli("freva")
    cli.parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    args = cli.parse_args(argv or sys.argv[1:])
    try:
        cli.run_cmd(args, **cli.kwargs)
    except KeyboardInterrupt:  # pragma: no cover
        print("KeyboardInterrupt, exiting", file=sys.stderr, flush=True)
        sys.exit(130)



"""
Created on 24.05.2016

@author: Sebastian Illing
"""
import os
from pathlib import Path
import pytest
import mock


def test_command(dummy_solr, capsys, temp_dir, admin_env):
    from evaluation_system.tests import run_cli
    from evaluation_system.model.solr import SolrFindFiles
    from evaluation_system.misc import config

    with mock.patch.dict(os.environ, admin_env, clear=True):
        with pytest.raises(SystemExit):
            run_cli(["solr", "foo"])
        assert "invalid choice" in capsys.readouterr().err
        run_cli(["solr", "index", dummy_solr.tmpdir])
        assert len(list(SolrFindFiles.search())) == 3
        assert len(list(SolrFindFiles.search(latest_version=False))) == 5
        run_cli(["solr", "index", dummy_solr.tmpdir, "--delete"])
        assert len(list(SolrFindFiles.search())) == 0
        assert len(list(SolrFindFiles.search(latest_version=False))) == 0


def test_forbidden_usage(capsys, dummy_env):
    from evaluation_system.tests import run_cli
    from freva.cli.admin import re_index, del_index

    with pytest.raises(SystemExit):
        run_cli(["solr", "index" "--help"])
    assert "invalid choice" in capsys.readouterr().err
    for func in (re_index, del_index):
        with pytest.raises(RuntimeError):
            func("foo")



"""Submit job to the sge workload manager."""
from __future__ import annotations
import logging

from .core import Job
from typing import ClassVar, Optional, Union

logger = logging.getLogger(__name__)


class SGEJob(Job):
    submit_command: ClassVar[str] = "qsub"
    cancel_command: ClassVar[str] = "qdel"
    config_name: ClassVar[str] = "sge"

    def __init__(
        self,
        name: Optional[str] = None,
        queue: Optional[str] = None,
        project: Optional[str] = None,
        resource_spec: Optional[str] = None,
        walltime: Optional[str] = None,
        job_extra: Optional[list[str]] = [],
        **base_class_kwargs,
    ):
        super().__init__(name=name, **base_class_kwargs)
        job_extra = job_extra or []
        header_lines = []
        if self.job_name is not None:
            header_lines.append("#$ -N %(job-name)s")
        if queue is not None:
            header_lines.append("#$ -q %(queue)s")
        if project is not None:
            header_lines.append("#$ -P %(project)s")
        if resource_spec is not None:
            header_lines.append("#$ -l %(resource_spec)s")
        if walltime is not None:
            header_lines.append("#$ -l h_rt=%(walltime)s")
        if self.log_directory is not None:
            header_lines.append("#$ -e %(log_directory)s/")
            header_lines.append("#$ -o %(log_directory)s/")
        header_lines.extend(["#$ -cwd", "#$ -j y"])
        header_lines.extend(["#$ %s" % arg for arg in job_extra])
        header_template = "\n".join(header_lines)

        config = {
            "job-name": self.job_name,
            "queue": queue,
            "project": project,
            "processes": self.worker_processes,
            "walltime": walltime,
            "resource_spec": resource_spec,
            "log_directory": self.log_directory,
        }
        self.job_header = header_template % config

        logger.debug("Job script: \n %s" % self.job_script())



"""
Created on 23.05.2016

@author: Sebastian Illing
"""
import gzip
import os
from pathlib import Path

import pytest


def test_ingest(dummy_solr):
    from evaluation_system.model.solr_core import SolrCore
    from evaluation_system.model.solr import SolrFindFiles
    from evaluation_system.misc.utils import supermakedirs

    latest_versions = [dummy_solr.files[0], dummy_solr.files[1], dummy_solr.files[3]]
    multiversion_latest = dummy_solr.files[3]
    old_versions = [dummy_solr.files[2], dummy_solr.files[4]]
    data_dir = Path(dummy_solr.tmpdir) / "cmip5"
    # test instances, check they are as expected
    SolrCore.load_fs(
        data_dir,
        abort_on_errors=True,
        core_all_files=dummy_solr.all_files,
        core_latest=dummy_solr.latest,
    )
    # check
    ff_all = SolrFindFiles(
        core="files", host=dummy_solr.solr_host, port=dummy_solr.solr_port
    )
    ff_latest = SolrFindFiles(
        core="latest", host=dummy_solr.solr_host, port=dummy_solr.solr_port
    )
    all_entries = [i for i in ff_all._search()]
    latest_entries = [i for i in ff_latest._search()]
    assert all([dummy_solr.tmpdir + "/" + e in all_entries for e in dummy_solr.files])
    assert all([dummy_solr.tmpdir + "/" + e in latest_entries for e in latest_versions])
    assert all(
        [dummy_solr.tmpdir + "/" + e not in latest_entries for e in old_versions]
    )

    # add new version
    new_version = (
        Path(dummy_solr.tmpdir)
        / "cmip5/output1/MOHC/HadCM3/decadal2009/mon/atmos/Amon/r7i2p1/v20120419/ua/ua_Amon_HadCM3_decadal2009_r7i2p1_200911-201912.nc"
    )
    new_version.parent.mkdir(parents=True, exist_ok=True)
    new_version.touch()
    SolrCore.load_fs(
        new_version,
        abort_on_errors=True,
        core_all_files=dummy_solr.all_files,
        core_latest=dummy_solr.latest,
    )
    assert set(ff_all._search()).symmetric_difference(set(all_entries)).pop() == str(
        new_version
    )
    assert (set(ff_latest._search()) - set(latest_entries)).pop() == str(new_version)
    # TODO: The below test does not make much sense, because data set versioning
    # doesn't really work, let's turn it off for now
    # assert (set(latest_entries) - set(ff_latest._search())).pop() == dummy_solr.tmpdir + '/' + multiversion_latest
    # test get_solr_fields (facets)
    facets = dummy_solr.all_files.get_solr_fields().keys()
    facets_to_be = [
        "model",
        "product",
        "realm",
        "version",
        "dataset",
        "institute",
        "file_name",
        "creation_time",
        "cmor_table",
        "time_frequency",
        "experiment",
        "timestamp",
        "file",
        "time",
        "variable",
        "_version_",
        "file_no_version",
        "project",
        "ensemble",
    ]
    assert sorted(facets) == sorted(facets_to_be)


def test_reload(dummy_solr):
    res = dummy_solr.all_files.reload()
    assert ["responseHeader"] == list(res.keys())


def test_unload_and_create(dummy_solr):

    res = dummy_solr.all_files.unload()
    status = dummy_solr.all_files.status()
    assert {} == status
    # with pytest.raises(FileNotFoundError):
    #    dummy_solr.all_files.create()
    dummy_solr.all_files.create(check_if_exist=False)
    assert len(dummy_solr.all_files.status()) >= 8



"""
.. moduleauthor:: Sebastian Illing / estani

This module manages the central configuration of the system.
"""
from __future__ import annotations
from typing import Optional, Sequence

import os
import os.path as osp
from pathlib import Path
import hashlib
import requests
from configparser import ConfigParser, NoSectionError, ExtendedInterpolation
import sys
import warnings
import toml

import appdirs
from evaluation_system.misc.utils import Struct, TemplateDict
from evaluation_system.misc import (
    logger as log,
    CONFIG_FILE,
    _DEFAULT_CONFIG_FILE_LOCATION,
)
from evaluation_system.misc.exceptions import ConfigurationException

DIRECTORY_STRUCTURE = Struct(LOCAL="local", CENTRAL="central", SCRATCH="scratch")
"""Type of directory structure that will be used to maintain state::

    local   := <home>/<base_dir>...
    central := <base_dir_location>/<base_dir>/<user>/...
    scratch := <base_dir_location>/<user>/<base_dir>...

We only use local at this time, but we'll be migrating to central in
the future for the next project phase."""
# Some defaults in case nothing is defined
USER_CONFIG_FILE_LOC = osp.join(
    appdirs.user_config_dir(), "freva", "evaluation_system.conf"
)
if Path(USER_CONFIG_FILE_LOC).exists():
    _DEFAULT_CONFIG_FILE_LOCATION = USER_CONFIG_FILE_LOC
EVALUATION_SYSTEM_HOME = (os.sep).join(osp.abspath(__file__).split(osp.sep)[:-4])
SPECIAL_VARIABLES = TemplateDict(EVALUATION_SYSTEM_HOME=EVALUATION_SYSTEM_HOME)

_DEFAULT_DRS_CONFIG_FILE = os.environ.get(
    "EVALUATION_SYSTEM_DRS_CONFIG_FILE",
    osp.abspath(osp.join(CONFIG_FILE, osp.pardir, "drs_config.toml")),
)
_PUBLIC_KEY_DIR = Path(CONFIG_FILE).parent
#: config options
BASE_DIR = "base_dir"
"The name of the directory storing the evaluation system (output, configuration, etc)"

ROOT_DIR = "root_dir"

DIRECTORY_STRUCTURE_TYPE = "directory_structure_type"
"""Defines which directory structure is going to be used.
See DIRECTORY_STRUCTURE"""

BASE_DIR_LOCATION = "base_dir_location"
"""The location of the directory defined in $base_dir."""

DATABASE_TIMEOUT = 2
"""Time out in seconds for the SQLite database"""

SCHEDULER_WAITING_TIME = 3
"""
Waiting time in seconds before the scheduler runs the job
We suggest: DATABASE_TIMEOUT+1
"""

SCRATCH_DIR = "scratch_dir"
""" The scratch dir of the user """


SCHEDULER_OUTPUT_DIR = "scheduler_output_dir"
""" Determines the folder the SLURM output files will be saved in """

SCHEDULER_INPUT_DIR = "scheduler_input_dir"
"""
Determines the folder the SLURM input files will be saved in.
This variable is  read by the User object,
use always user.getUserSchedulerInputDir()!
"""

# SCHEDULER_COMMAND='/client/bin/sbatch'
""" The command for the scheduler """

SCHEDULER_OPTIONS = "--begin=now+" + str(SCHEDULER_WAITING_TIME)

NUMBER_OF_PROCESSES = 24

DATABASE_FILE = "database_file"
""" Determines the path of the database file """

# config file section
CONFIG_SECTION_NAME = "evaluation_system"
"""This is the name of the section in the configuration file where
the central configuration is being stored"""

#: Plugins *******************
PLUGINS = "plugin:"
"""Sections that start with this prefix are meant for configuring the plug-ins.
This keyword is also used for retrieving all plug-ins configurations."""

PLUGIN_PATH = "plugin_path"
""""Path to the plug-in's home"""

PLUGIN_PYTHON_PATH = "python_path"
"""Path to the plug-in's python sources that should be added to
python's path when being run."""

PLUGIN_MODULE = "module"
"""The full qualified module name where the plug-in is implemented."""

PREVIEW_PATH = "preview_path"
"""path to preview pictures."""

GIT_BASH_STARTOPTIONS = "git_bash_startoptions"
"""We use -lc, use this option if something else is required"""

#: database #####################
DB_HOST = "db.host"
""" name of the database server """

DB_USER = "db.user"
""" database user """

DB_PASSWD = "db.passwd"
""" database password """

DB_DB = "db.db"
""" the database name on the server """

DB_PORT = "db.port"
""" database connection port"""

#: Solr #########################
SOLR_HOST = "solr.host"
"""Hostname of the Solr instance."""

SOLR_PORT = "solr.port"
"""Port number of the Solr instance."""

SOLR_CORE = "solr.core"
"""Core name of the Solr instance."""


_config = None
_drs_config = None

exclude: list[str] = ["extra_scheduler_options"]
"""This is a list of strings that excludes entries from being saved to
the history database entries."""


def _get_public_key(project_name: str) -> str:
    key_file = os.environ.get("PUBKEY", None) or _PUBLIC_KEY_DIR / f"{project_name}.crt"
    sha = ""
    try:
        with Path(key_file).open() as f:
            key = "".join([k.strip() for k in f.readlines() if not k.startswith("-")])
        sha = hashlib.sha512(key.encode()).hexdigest()
    except FileNotFoundError:
        warnings.warn(
            (
                f"{key_file} not found. Secrets are stored in central vault and a"
                "public key is needed to open the vault. Without the public key"
                " you won't be probaply be able to establish as database "
                "connection."
            ),
            category=Warning,
        )
    return sha


def _read_secrets(
    sha: str,
    key: str,
    *db_hosts: Sequence[str],
    port: int = 5002,
    protocol: str = "http",
) -> Optional[str]:
    """Query the vault for data database secrets, of a given key."""
    for db_host in db_hosts:
        url = f"{protocol}://{db_host}:{port}/vault/data/{sha}"
        try:
            req = requests.get(url).json()
        except (requests.exceptions.ConnectionError, requests.exceptions.InvalidURL):
            req = {}
        try:
            return req[key]
        except KeyError:
            pass
    return None


def reloadConfiguration() -> None:
    """Reloads the configuration.
    This can be used for reloading a new configuration from disk.
    At the present time it has no use other than setting different configurations
    for testing, since the framework is restarted every time an analysis is
    performed."""
    global _config
    _config = {
        BASE_DIR: "evaluation_system",
        BASE_DIR_LOCATION: os.path.expanduser("~"),
        DIRECTORY_STRUCTURE_TYPE: DIRECTORY_STRUCTURE.LOCAL,
        PLUGINS: {},
    }

    config_file = os.environ.get(
        "EVALUATION_SYSTEM_CONFIG_FILE", _DEFAULT_CONFIG_FILE_LOCATION
    )
    log.debug("Loading configuration file from: %s" % config_file)
    if config_file and os.path.isfile(config_file):
        config_parser = ConfigParser(interpolation=ExtendedInterpolation())
        with open(config_file, "r") as fp:
            config_parser.read_file(fp)
            if not config_parser.has_section(CONFIG_SECTION_NAME):
                raise ConfigurationException(
                    (
                        "Configuration file is missing section %s.\n"
                        + "For Example:\n[%s]\nprop=value\n..."
                    )
                    % (CONFIG_SECTION_NAME, CONFIG_SECTION_NAME)
                )
            else:
                _config.update(config_parser.items(CONFIG_SECTION_NAME))
                for plugin_section in [
                    s for s in config_parser.sections() if s.startswith(PLUGINS)
                ]:
                    _config[PLUGINS][
                        plugin_section[len(PLUGINS) :]
                    ] = SPECIAL_VARIABLES.substitute(
                        dict(config_parser.items(plugin_section))
                    )

                db_hosts = (
                    config_parser[CONFIG_SECTION_NAME]["db.host"],
                    config_parser[CONFIG_SECTION_NAME]["project_name"] + "_vault",
                )
                # This will look first for secrets set in the config file. It will only
                # load the public key for the vault if any are missing and it will only
                # look for the missing keys in the vault. Any keys set in the config
                # will take priority
                secret_store_keys: list[str] = []
                for secret in ("db.user", "db.passwd", "db.db", "db.host"):
                    value = _config.get(secret, None)
                    if value:
                        _config[secret] = value
                    else:
                        secret_store_keys.append(secret)
                if len(secret_store_keys) > 0:
                    secret_store_keys += ["db.port"]
                    sha: str = _get_public_key(
                        config_parser[CONFIG_SECTION_NAME]["project_name"]
                    )
                    for secret in secret_store_keys:
                        _config[secret] = _read_secrets(sha, secret, *db_hosts)
            log.debug("Configuration loaded from %s", config_file)
    else:
        log.debug(
            "No configuration file found in %s. Using default values.", config_file
        )

    _config = SPECIAL_VARIABLES.substitute(_config, recursive=False)
    # perform all special checks
    if not DIRECTORY_STRUCTURE.validate(_config[DIRECTORY_STRUCTURE_TYPE]):
        raise ConfigurationException(
            "value (%s) of %s is not valid. Should be one of: %s"
            % (
                _config[DIRECTORY_STRUCTURE_TYPE],
                DIRECTORY_STRUCTURE_TYPE,
                ", ".join(DIRECTORY_STRUCTURE.toDict().values()),
            )
        )


_nothing = object()


def get(config_prop, default=_nothing):
    """Returns the value stored for the given config_prop.
    If the config_prop is not found and no default value is provided an exception
    will be thrown. If not the default value is returned.

    :param config_prop: property for which it's value is looked for.
    :type config_prop: str
    :param default: If the property is not found this value is returned.
    :return: the value associated with the given property, the default one
    if not found or an exception is thrown if no default is provided."""
    if config_prop in _config:
        return _config[config_prop]
    elif default != _nothing:
        return default
    else:
        raise ConfigurationException("No configuration for %s" % config_prop)


def get_plugin(plugin_name, config_prop, default=_nothing):
    """Returns the value stored for the given config_prop at the given plugin.
    If the config_prop is not found and no default value is provided an exception
    will be thrown. If not the default value is returned. If the plug.in name does
    not exists an exception is thrown.

    :param plugin_name: name of an existing plug-in as returned by
    :class:`get`(:class:`PLUGINS`)
    :type plugin_name: str
    :param config_prop: property for which it's value is looked for.
    :type config_prop: str
    :param default: If the property is not found this value is returned.
    :return: the value associated with the given property, the default one
    if not found
        or an exception is thrown if no default is provided."""
    if plugin_name in _config[PLUGINS]:
        _plugin_config = _config[PLUGINS][plugin_name]
        if config_prop in _plugin_config:
            return _plugin_config[config_prop]
        elif default != _nothing:
            return default
        else:
            raise ConfigurationException(
                "No configuration for %s for plug-in %s" % (config_prop, plugin_name)
            )
    else:
        raise ConfigurationException("No plug-in named %s" % plugin_name)


def keys():
    """Returns all the keys from the current configuration."""
    return _config.keys()


def get_section(section_name, config_file=None):
    conf = ConfigParser(interpolation=ExtendedInterpolation())

    config_file = config_file or os.environ.get(
        "EVALUATION_SYSTEM_CONFIG_FILE", _DEFAULT_CONFIG_FILE_LOCATION
    )
    conf.read(config_file)
    try:
        section = dict(conf.items(section_name))
    except NoSectionError:
        raise NoSectionError(f'There is no "{section_name}" section in config file')
    return SPECIAL_VARIABLES.substitute(section)


def get_drs_config():
    global _drs_config
    if _drs_config is None:
        drs_config = os.environ.get(
            "EVALUATION_SYSTEM_DRS_CONFIG_FILE", str(_DEFAULT_DRS_CONFIG_FILE)
        )
        with open(drs_config, "r") as drs_file:
            _drs_config = toml.load(drs_file)
    return _drs_config
